"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 1, 2011.
"""
from everest.repositories.rdb.utils import mapper
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityGrandchild
from everest.tests.complete_app.entities import MyEntityParent
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy.orm import relationship

__docformat__ = 'reStructuredText en'
__all__ = ['create_metadata',
           ]



def create_metadata(engine):
    metadata = MetaData()
    #
    # TABLES
    #
    my_entity_parent_tbl = \
        Table('my_entity_parent', metadata,
              Column('my_entity_parent_id', Integer, primary_key=True),
              Column('text', String),
              Column('text_ent', String),
              )
    # 1:1 MyEntity <=> MyEntityParent
    my_entity_tbl = \
        Table('my_entity', metadata,
              Column('my_entity_id', Integer, primary_key=True),
              Column('text', String),
              Column('text_ent', String),
              Column('number', Integer),
              Column('my_entity_parent_id', Integer,
                     ForeignKey(my_entity_parent_tbl.c.my_entity_parent_id),
                     nullable=False),
              )
    # 1:n MyEntity <-> MyEntityChild
    my_entity_child_tbl = \
        Table('my_entity_child', metadata,
              Column('text', String),
              Column('text_ent', String),
              Column('my_entity_child_id', Integer, primary_key=True),
              Column('my_entity_id', Integer,
                     ForeignKey(my_entity_tbl.c.my_entity_id),
                     nullable=False),
              )
    # n:m MyEntityChild <-> MyEntityGrandchild
    my_entity_grandchild_tbl = \
        Table('my_entity_grandchild', metadata,
              Column('text', String),
              Column('text_ent', String),
              Column('my_entity_grandchild_id', Integer, primary_key=True),
              )
    my_entity_child_children_tbl = \
        Table('my_entity_child_children', metadata,
              Column('my_entity_child_id', Integer,
                     ForeignKey(my_entity_child_tbl.c.my_entity_child_id),
                     nullable=False),
              Column('my_entity_grandchild_id', Integer,
                     ForeignKey(
                        my_entity_grandchild_tbl.c.my_entity_grandchild_id),
                     nullable=False)
              )
    #
    # MAPPERS
    #

    mapper(MyEntityParent, my_entity_parent_tbl,
           id_attribute='my_entity_parent_id',
           properties=
            dict(child=relationship(MyEntity,
                                    uselist=False,
                                    back_populates='parent'),
                 )
           )
    mapper(MyEntity, my_entity_tbl,
           id_attribute='my_entity_id',
           properties=
            dict(parent=relationship(MyEntityParent,
                                     uselist=False,
                                     back_populates='child'),
                 children=relationship(MyEntityChild,
                                       back_populates='parent',
                                       cascade="all, delete-orphan"),
                 )
           )
    mapper(MyEntityChild, my_entity_child_tbl,
           id_attribute='my_entity_child_id',
           properties=
            dict(parent=relationship(MyEntity,
                                     uselist=False,
                                     back_populates='children',
                                     cascade='save-update'
                                     ),
                 children=
                    relationship(MyEntityGrandchild,
                                 secondary=my_entity_child_children_tbl,
                                 back_populates='parent'),
                 ),
           )
    mapper(MyEntityGrandchild, my_entity_grandchild_tbl,
           id_attribute='my_entity_grandchild_id',
           properties=
            dict(parent=relationship(MyEntityChild,
                                     uselist=False,
                                     secondary=my_entity_child_children_tbl,
                                     back_populates='children'),
                 ),
           )
    # Create the mappers.
#    metadata.bind = engine
    metadata.create_all(engine)
    return metadata



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 5, 2011.
"""
from everest.repositories.rdb import SqlFilterSpecificationVisitor
from everest.querying.specifications import ValueEqualToFilterSpecification
from everest.testing import Pep8CompliantTestCase
from everest.tests.simple_app.entities import FooEntity

__docformat__ = 'reStructuredText en'
__all__ = ['SqlFilterSpecificationVisitorTestCase',
           ]


class SqlFilterSpecificationVisitorTestCase(Pep8CompliantTestCase):
    def test_custom_clause(self):
        obj = object()
        func = lambda value: obj
        spec = ValueEqualToFilterSpecification('foo', 'bar')
        factory_map = {('foo', spec.operator.name):func}
        visitor = SqlFilterSpecificationVisitor(FooEntity,
                                                custom_clause_factories=
                                                                 factory_map)
        visitor.visit_nullary(spec)
        self.assert_true(visitor.expression is obj)



"""
Resource related utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 3, 2011.
"""
from pyramid.compat import string_types
from pyramid.compat import urlparse
from pyramid.threadlocal import get_current_registry
from pyramid.threadlocal import get_current_request
from pyramid.traversal import model_path

from everest.interfaces import IResourceUrlConverter
from everest.repositories.utils import as_repository
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IRelation
from everest.resources.interfaces import IResource
from everest.resources.interfaces import IService
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface.interfaces import IInterface # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['as_member',
           'get_collection_class',
           'get_member_class',
           'get_registered_collection_resources',
           'get_resource_class_for_relation',
           'get_resource_url',
           'get_root_collection',
           'get_service',
           'is_resource_url',
           'provides_collection_resource',
           'provides_member_resource',
           'provides_resource',
           'resource_to_url',
           'url_to_resource',
           ]


def get_root_collection(resource):
    """
    Returns a clone of the collection from the repository registered for the
    given registered resource.

    :param resource: registered resource
    :type resource: class implementing or instance providing or subclass of
        a registered resource interface.
    """
    repo = as_repository(resource)
    return repo.get_collection(resource)


def get_member_class(resource):
    """
    Returns the registered member class for the given resource.

    :param resource: registered resource
    :type resource: class implementing or instance providing or subclass of
        a registered resource interface.
    """
    reg = get_current_registry()
    if IInterface in provided_by(resource):
        member_class = reg.getUtility(resource, name='member-class')
    else:
        member_class = reg.getAdapter(resource, IMemberResource,
                                      name='member-class')
    return member_class


def get_collection_class(resource):
    """
    Returns the registered collection resource class for the given marker
    interface or member resource class or instance.

    :param rc: registered resource
    :type rc: class implementing or instance providing or subclass of
        a registered resource interface.
    """
    reg = get_current_registry()
    if IInterface in provided_by(resource):
        coll_class = reg.getUtility(resource, name='collection-class')
    else:
        coll_class = reg.getAdapter(resource, ICollectionResource,
                                    name='collection-class')
    return coll_class


def get_resource_class_for_relation(relation):
    """
    Returns the resource class that was registered for the given
    relation.

    :param str relation: relation string.
    """
    reg = get_current_registry()
    return reg.getUtility(IRelation, name=relation)


def as_member(entity, parent=None):
    """
    Adapts an object to a location aware member resource.

    :param entity: a domain object for which a resource adapter has been
        registered
    :type entity: an object implementing
        :class:`everest.entities.interfaces.IEntity`
    :param parent: optional parent collection resource to make the new member
        a child of
    :type parent: an object implementing
        :class:`everest.resources.interfaces.ICollectionResource`
    :returns: an object implementing
        :class:`everest.resources.interfaces.IMemberResource`
    """
    reg = get_current_registry()
    rc = reg.getAdapter(entity, IMemberResource)
    if not parent is None:
        rc.__parent__ = parent # interface method pylint: disable=E1121
    return rc


def is_resource_url(url_string):
    """
    Checks if the given URL string is a resource URL.

    Currently, this check only looks if the URL scheme is either "http" or
    "https".
    """
    return isinstance(url_string, string_types) \
           and urlparse.urlparse(url_string).scheme in ('http', 'https') # pylint: disable=E1101


def get_resource_url(resource):
    """
    Returns the URL for the given resource.
    """
    path = model_path(resource)
    parsed = list(urlparse.urlparse(path))
    parsed[1] = ""
    return urlparse.urlunparse(parsed)


def provides_resource(obj):
    """
    Checks if the given type or instance provides the
    :class:`everest.resources.interfaces.IResource` interface.
    """
    if isinstance(obj, type):
        obj = object.__new__(obj)
    return IResource in provided_by(obj)


def provides_member_resource(obj):
    """
    Checks if the given type or instance provides the
    :class:`everest.resources.interfaces.IMemberResource` interface.
    """
    if isinstance(obj, type):
        obj = object.__new__(obj)
    return IMemberResource in provided_by(obj)


def provides_collection_resource(obj):
    """
    Checks if the given type or instance provides the
    :class:`everest.resources.interfaces.ICollectionResource` interface.
    """
    if isinstance(obj, type):
        obj = object.__new__(obj)
    return ICollectionResource in provided_by(obj)


def get_registered_collection_resources():
    """
    Returns a list of all registered collection resource classes.
    """
    reg = get_current_registry()
    return [util.component
            for util in reg.registeredUtilities()
            if util.name == 'collection-class']


def get_service():
    """
    Registers the object registered as the service utility.

    :returns: object implementing
        :class:`everest.interfaces.IService`
    """
    reg = get_current_registry()
    return reg.getUtility(IService)


def resource_to_url(resource, request=None, quote=False):
    """
    Converts the given resource to a URL.

    :param request: Request object (required for the host name part of the
      URL). If this is not given, the current request is used.
    :param bool quote: If set, the URL returned will be quoted.
    """
    if request is None:
        request = get_current_request()
#    cnv = request.registry.getAdapter(request, IResourceUrlConverter)
    reg = get_current_registry()
    cnv = reg.getAdapter(request, IResourceUrlConverter)
    return cnv.resource_to_url(resource, quote=quote)


def url_to_resource(url, request=None):
    """
    Converts the given URL to a resource.

    :param request: Request object (required for the host name part of the
      URL). If this is not given, the current request is used.
    """
    if request is None:
        request = get_current_request()
#    cnv = request.registry.getAdapter(request, IResourceUrlConverter)
    reg = get_current_registry()
    cnv = reg.getAdapter(request, IResourceUrlConverter)
    return cnv.url_to_resource(url)



"""
Representer configuration.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 8, 2012.
"""
from collections import defaultdict
from pyramid.compat import iteritems_
from pyramid.compat import string_types

__docformat__ = 'reStructuredText en'
__all__ = ['RepresenterConfiguration',
           ]

# Configuration option name declarations.

_ATTRIBUTES_CONFIG_OPTION = 'attributes'
IGNORE_OPTION = 'ignore'
WRITE_AS_LINK_OPTION = 'write_as_link'
WRITE_MEMBERS_AS_LINK_OPTION = 'write_members_as_link'
REPR_NAME_OPTION = 'repr_name'


class RepresenterConfiguration(object):
    """
    Class maintaining representer configuration options.

    At present, this can also be used as base class for containers declaring
    representer configuration data as static class attributes. However, this
    usage is discouraged and should be replaced with the new ZCML based
    declaration syntax.

    Representer configuration objects maintain two kinds of configuration
    data:

    1) Generic options. These can be any key:value pairs. Derived
       classes need to declare valid options in the
       :cvar:`_default_config_options` class variable.

    2) Attributes options. These are kept in a dictionary mapping the mapped
       attribute name to a dictionary of options which control the way each
       attribute is mapped. Valid option names for a given attribute are:

       %(REPR_NAME_OPTION)s :
         The name to use for this attribute in the representation.
       %(WRITE_AS_LINK_OPTION)s :
         Write this mapped attribute as a link rather than as a full
         representation.
       %(WRITE_MEMBERS_AS_LINK_OPTION)s :
         Write members of a mapped collection attribute as a link rather
         than as a full representation.
       %(IGNORE_OPTION)s :
         Ignore this attribute when creating a representation.

       Derived classes may add more allowed mapping options; those must be
       declared in the :cvar:`_default_attributes_options` class variable.
    """ % globals() # doc string must not be assigned pylint: disable=W0106
    #: Default configuration option names (immutable).
    _default_config_options = {}
    #: Default mapping option names (immutable).
    _default_attributes_options = {IGNORE_OPTION:None,
                                   WRITE_AS_LINK_OPTION:None,
                                   WRITE_MEMBERS_AS_LINK_OPTION:None,
                                   REPR_NAME_OPTION:None}

    def __init__(self, options=None, attribute_options=None):
        # {generic config option name : option value}
        self.__options = self._default_config_options.copy()
        # {attr key : { attr name : {{option name : option value}}}
        self.__attribute_options = \
                        defaultdict(self._default_attributes_options.copy)
        self.__update(options, attribute_options)

    def copy(self):
        """
        Return a copy of this configuration.
        """
        return self.__class__(options=self.__options,
                              attribute_options=self.__attribute_options)

    def update(self, configuration):
        """
        Update this configuration with the given other configuration. Only
        not-None values will be overridden.
        """
        self.__update(configuration.get_options(),
                      configuration.get_attribute_options())

    def get_option(self, name):
        """
        Returns the value for the specified generic configuration option.

        :returns: configuration option value or `None`, if the option was not
          set.
        """
        self.__validate_option_name(name)
        return self.__options.get(name, None)

    def set_option(self, name, value):
        """
        Sets the specified generic configuration option to the given value.
        """
        self.__validate_option_name(name)
        self.__options[name] = value

    def get_options(self):
        """
        Returns a copy of the generic configuration options.
        """
        return self.__options.copy()

    def set_attribute_option(self, attribute, option_name, option_value):
        """
        Sets the given attribute option to the given value for the specified
        attribute.
        """
        self.__validate_attribute_option_name(option_name)
        attribute_key = self.__make_key(attribute)
        mp_options = self.__attribute_options.setdefault(attribute_key, {})
        mp_options[option_name] = option_value

    def get_attribute_option(self, attribute, option_name):
        """
        Returns the value of the given attribute option for the specified
        attribute.
        """
        self.__validate_attribute_option_name(option_name)
        attribute_key = self.__make_key(attribute)
        return self.__attribute_options[attribute_key].get(option_name)

    def get_attribute_options(self, attribute=None):
        """
        Returns a copy of the mapping options for the given attribute name
        or a copy of all mapping options, if no attribute name is provided.
        All options that were not explicitly configured are given a default
        value of `None`.

        :param tuple attribute_key: attribute name or tuple specifying an
          attribute path.
        :returns: mapping options dictionary (including default `None` values)
        """
        attribute_key = self.__make_key(attribute)
        if attribute_key is None:
            opts = defaultdict(self._default_attributes_options.copy)
            for attr, mp_options in iteritems_(self.__attribute_options):
                opts[attr].update(mp_options)
        else:
            opts = self._default_attributes_options.copy()
            attr_opts = self.__attribute_options[attribute_key]
            opts.update(attr_opts)
        return opts

    def __make_key(self, attribute):
        if isinstance(attribute, string_types):
            key = tuple(attribute.split('.'))
        elif not (isinstance(attribute, tuple) or attribute is None):
            key = tuple(attribute)
        else:
            key = attribute
        return key

    def __update(self, opts, mp_opts):
        if not opts is None:
            for option_name, option_value in iteritems_(opts):
                if not option_value is None:
                    self.set_option(option_name, option_value)
        if not mp_opts is None:
            for attr_name, attr_mp_options in iteritems_(mp_opts):
                for mp_opt_name, mp_opt_value in iteritems_(attr_mp_options):
                    if not mp_opt_value is None:
                        self.set_attribute_option(attr_name,
                                                mp_opt_name, mp_opt_value)

    def __validate_option_name(self, name):
        if not (name in self._default_config_options.keys()
                or name == _ATTRIBUTES_CONFIG_OPTION):
            raise ValueError('Invalid configuration option name "%s" for '
                             '%s representer.' %
                             (name, self.__class__.__name__))

    def __validate_attribute_option_name(self, name):
        if not name in self._default_attributes_options.keys():
            raise ValueError('Invalid attribute option name "%s" '
                             'for %s representer.'
                             % (name, self.__class__.__name__))



"""
Renderers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created Oct 7, 2011.
"""
from pyramid.interfaces import IRenderer

from everest.mime import AtomMime
from everest.mime import CsvMime
from everest.mime import JsonMime
from everest.mime import XmlMime
from everest.representers.utils import as_representer
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IResource
from zope.interface import implementer # pylint: disable=E0611,F0401
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401


__docformat__ = "reStructuredText en"
__all__ = ['AtomRenderer',
           'CsvRenderer',
           'JsonRenderer',
           'RendererFactory',
           'ResourceRenderer',
           'XmlRenderer',
           ]


class RendererFactory(object):
    """
    Factory for renderers based on names.
    """
    def __init__(self, info):
        self.__name = info.name

    def __call__(self, value, system):
        if self.__name == 'csv':
            rnd = CsvRenderer()
        elif self.__name == 'json':
            rnd = JsonRenderer()
        elif self.__name == 'xml':
            rnd = XmlRenderer()
        elif self.__name == 'atom':
            rnd = AtomRenderer()
        else:
            raise ValueError('Unknown renderer name "%s"' % self.__name)
        return rnd(value, system)


@implementer(IRenderer)
class ResourceRenderer(object):
    """
    Renderer for resources.

    Uses a representer to perform the resource -> representation conversion.
    """
    def __init__(self, content_type): # redef format pylint:disable=W0622
        self._content_type = content_type

    def __call__(self, value, system):
        context = value.get('context', system.get('context'))
        if not self._validate(context):
            raise ValueError('Invalid representation.')
        self._prepare_response(system)
        # Assemble response.
        rpr = as_representer(context, self._content_type)
        return rpr.to_bytes(context)

    @property
    def _format(self):
        return self._content_type.mime_type_string

    def _validate(self, value):
        return IResource in  provided_by(value)

    def _prepare_response(self, system):
        # Set up response type.
        rsp = system['request'].response
        rsp.content_type = self._format


class CsvRenderer(ResourceRenderer):
    """
    Renderer creating a CSV representation from a resource.
    """
    def __init__(self):
        ResourceRenderer.__init__(self, CsvMime)

    def _prepare_response(self, system):
        ResourceRenderer._prepare_response(self, system)
        context = system['context']
        if ICollectionResource in provided_by(context):
            # Disable batching for CSV rendering.
            context.slice = None


class JsonRenderer(ResourceRenderer):
    """
    Renderer creating a JSON representation from a resource.
    """
    def __init__(self):
        ResourceRenderer.__init__(self, JsonMime)


class XmlRenderer(ResourceRenderer):
    """
    Renderer creating a XML representation from a resource.
    """
    def __init__(self):
        ResourceRenderer.__init__(self, XmlMime)


class AtomRenderer(ResourceRenderer):
    """
    Renderer creating an ATOM representation from a resource.
    """
    def __init__(self):
        ResourceRenderer.__init__(self, AtomMime)



"""
WSGI filter to help a Flex client interact with an everest server.

Background:

The Flash browser plugin is not able to

 1. Properly handle response messages for which the status code is <> 200.
    This makes it impossible to properly handle errors on the client side
    (e.g. distinguish a bad request form an internal server error); and
 2. Set Accept headers for GET requests. This causes the default Flex
    Accept header to be sent ::
        text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
    which prompts evererst to respond with XML, not with ATOM.

Solution:

On ingress, the filter replaces for GET calls from a Flex client the
Accept header with 'application/atom+xml'.

On egress, this filter processes all response messages for calls from a
Flex client with status code <> 200 by a) setting the status code to
'200 OK' and b) wrapping the error message n the payload like this ::
     <error>
      <code>400</code>
      <message>Bad request</message>
      <details>...</details>
    </error>

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Sep 27, 2011.
"""
from everest.constants import RequestMethods
from everest.mime import AtomMime
from lxml import etree
from pyramid.compat import string_types
from pyramid.httpexceptions import HTTPMovedPermanently
from pyramid.httpexceptions import HTTPOk
from pyramid.httpexceptions import HTTPTemporaryRedirect
from pyramid.httpexceptions import HTTPUnauthorized
from xml.sax.saxutils import escape
import re

__docformat__ = 'reStructuredText en'
__all__ = ['FlexFilter',
           ]


# FIXME: The following code is taken from Paste 1.7.5 and wsgifilter to
#        avoid a dependency on Paste which has not been ported to Python 3.x

def header_value(headers, name):
    name = name.lower()
    result = [value for header, value in headers
              if header.lower() == name]
    if result:
        return ','.join(result)
    else:
        return None


def replace_header(headers, name, value):
    name = name.lower()
    i = 0
    result = None
    while i < len(headers):
        if headers[i][0].lower() == name:
            assert not result, "two values for the header '%s' found" % name
            result = headers[i][1]
            headers[i] = (name, value)
        i += 1
    if not result:
        headers.append((name, value))
    return result


class Filter(object):
    """
    Class that implements WSGI output-filtering middleware
    """

    # If this is true, then conditional requests will be diabled
    # (e.g., If-Modified-Since)
    force_no_conditional = True

    conditional_headers = [
        'HTTP_IF_MODIFIED_SINCE',
        'HTTP_IF_NONE_MATCH',
        'HTTP_ACCEPT_ENCODING',
        ]

    # If true, then any status code will be filtered; otherwise only
    # 200 OK responses are filtered
    filter_all_status = False

    # If you provide this (a string or list of string mimetypes) then
    # only content with this mimetype will be filtered
    filter_content_types = ('text/html',)

    # If this is set, then HTTPEncode will be used to decode the value
    # given provided mimetype and this output
    format_output = None

    # You can also use a specific format object, which forces the
    # parsing with that format
    format = None

    # If you aren't using a format but you want unicode instead of
    # 8-bit strings, then set this to true
    decode_unicode = False

    # When we get unicode back from the filter, we'll use this
    # encoding and update the Content-Type:
    output_encoding = 'utf8'

    def __init__(self, app):
        self.app = app
        if isinstance(self.format, string_types):
            from httpencode import get_format
            self.format = get_format(self.format)
        if (self.format is not None
            and self.filter_content_types is Filter.filter_content_types):
            self.filter_content_types = self.format.content_types

    def __call__(self, environ, start_response):
        if self.force_no_conditional:
            for key in self.conditional_headers:
                if key in environ:
                    del environ[key]
        # @@: I should actually figure out a way to deal with some
        # encodings, particular since stuff we don't care about like
        # text/javascript could be gzipped usefully.
        if 'HTTP_ACCEPT_ENCODING' in environ:
            del environ['HTTP_ACCEPT_ENCODING']
        shortcutted = []
        captured = []
        written_output = []
        def replacement_start_response(status, headers, exc_info=None):
            if not self.should_filter(status, headers, exc_info):
                shortcutted.append(None)
                return start_response(status, headers, exc_info)
            if exc_info is not None and shortcutted:
                raise exc_info[0], exc_info[1], exc_info[2]
            # Otherwise we don't care about exc_info...
            captured[:] = [status, headers]
            return written_output.append
        app_iter = self.app(environ, replacement_start_response)
        if shortcutted:
            # We chose not to filter
            return app_iter
        if not captured or written_output:
            # This app hasn't called start_response We can't do
            # anything magic with it; or it used the start_response
            # writer, and we still can't do anything with it
            try:
                for chunk in app_iter:
                    written_output.append(chunk)
            finally:
                if hasattr(app_iter, 'close'):
                    app_iter.close()
            app_iter = written_output
        try:
            return self.filter_output(
                environ, start_response,
                captured[0], captured[1], app_iter)
        finally:
            if hasattr(app_iter, 'close'):
                app_iter.close() # pylint: disable=E1103

    def paste_deploy_middleware(cls, app, global_conf, **app_conf): # pylint: disable=W0613
        # You may wish to override this to make it convert the
        # arguments or use global_conf.  To declare your entry
        # point use:
        # setup(
        #   entry_points="""
        #   [paste.filter_app_factory]
        #   myfilter = myfilter:MyFilter.paste_deploy_middleware
        #   """)
        return cls(app, **app_conf)

    paste_deploy_middleware = classmethod(paste_deploy_middleware)

    def should_filter(self, status, headers, exc_info): # pylint: disable=W0613
        if not self.filter_all_status:
            if not status.startswith('200'):
                return False
        content_type = header_value(headers, 'content-type')
        if content_type and ';' in content_type:
            content_type = content_type.split(';', 1)[0]
        if content_type in self.filter_content_types:
            return True
        return False

    _charset_re = re.compile(
        r'charset="?([a-z0-9-_.]+)"?', re.I)

    # @@: I should do something with these:
    #_meta_equiv_type_re = re.compile(
    #    r'<meta[^>]+http-equiv="?content-type"[^>]*>', re.I)
    #_meta_equiv_value_re = re.compile(
    #    r'value="?[^">]*"?', re.I)

    def filter_output(self, environ, start_response,
                      status, headers, app_iter):
        content_type = header_value(headers, 'content-type')
        if ';' in content_type:
            content_type = content_type.split(';', 1)[0]
        if self.format_output:
            import httpencode
            format = httpencode.registry.find_format_match(self.format_output, content_type) # pylint: disable=W0622
        else:
            format = self.format
        if format:
            data = format.parse_wsgi_response(status, headers, app_iter)
        else:
            data = ''.join(app_iter)
            if self.decode_unicode:
                # @@: Need to calculate encoding properly
                full_ct = header_value(headers, 'content-type') or ''
                match = self._charset_re.search(full_ct)
                if match:
                    encoding = match.group(1)
                else:
                    # @@: Obviously not a great guess
                    encoding = 'utf8'
                data = data.decode(encoding, 'replace')
        new_output = self.filter(
            environ, headers, data)
        if format:
            app = format.responder(new_output, headers=headers)
            app_iter = app(environ, start_response)
            return app_iter
        else:
            enc_data = []
            encoding = self.output_encoding
            if not isinstance(new_output, string_types):
                for chunk in new_output:
                    if isinstance(chunk, unicode):
                        chunk = chunk.encode(encoding)
                    enc_data.append(chunk)
            elif isinstance(new_output, unicode):
                enc_data.append(new_output.encode(encoding))
            else:
                enc_data.append(new_output)
            start_response(status, headers)
            return enc_data

    def filter(self, environ, headers, data):
        raise NotImplementedError


class FlexFilter(Filter):

    def __call__(self, environ, start_response):
        if self.__detect_flex(environ) \
           and environ['REQUEST_METHOD'] == RequestMethods.GET \
           and environ.get('HTTP_ACCEPT'):
            # Override the Accept header. We prefer ATOM, but also allow
            # the server to decide if ATOM is not available.
            environ['HTTP_ACCEPT'] = \
                 "%s;q=0.9,*/*;q=0.8" % AtomMime.mime_type_string
        return Filter.__call__(self, environ, start_response)

    def should_filter(self, status, headers, exc_info):
        return not (status.startswith('2')
                    or status == HTTPUnauthorized().status
                    or status == HTTPMovedPermanently().status)

    def filter_output(self, environ, start_response,
                      status, headers, app_iter):
        if self.__detect_flex(environ):
            # All the fuzz is just for Flex/Flash clients.
            # Wrap start response to pass HTTP OK back to Flash. Also, we
            # need to have access to the status later.
            environ['flexfilter.status'] = status
            wrap_start_response = \
              lambda status, headers: start_response(HTTPOk().status, headers)
            return Filter.filter_output(self, environ, wrap_start_response,
                                        status, headers, app_iter)
        else:
            #unfiltered response for non flash clients
            start_response(status, headers)
            return app_iter

    def filter(self, environ, headers, data):
        status = environ.pop('flexfilter.status')
        if status == HTTPTemporaryRedirect().status:
            response = '''
                        <error>
                          <code>%s</code>
                          <location>%s</location>
                          <message>%s</message>
                        </error>
                        ''' % (status, header_value(headers, 'Location'),
                               header_value(headers, 'Warning'))
            replace_header(headers, 'Content-Length', len(response))
        else:
            root = etree.HTML(data)
            message = escape(etree.tostring(root.find('.//body'),
                                            method="text").strip())
            if not message:
                message = root.find('.//title').text
            details = ""
            code_node = root.find('.//code')
            if code_node  is not None and code_node.text  is not None:
                details = escape(code_node.text)
                # Shorten a bit.
                pos = details.find(',')
                if pos != -1:
                    details = details[:pos]
            response = '''
                        <error>
                          <code>%s</code>
                          <message>%s</message>
                          <details>%s</details>
                        </error>
                        ''' % (status, message, details)
            replace_header(headers, 'Content-Length', len(response))
        return response

    def __detect_flex(self, environ):
        query_string = environ.get('QUERY_STRING')
        referer = environ.get('HTTP_REFERER')
        return (referer and '.swf' in referer) \
               or (query_string and 'flashfilter=true' in query_string)



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from everest.interfaces import IMime
from everest.mime import XmlMime
from everest.mime import get_registered_mime_strings
from everest.mime import get_registered_mime_type_for_extension
from everest.mime import get_registered_mime_type_for_name
from everest.mime import get_registered_mime_type_for_string
from everest.mime import get_registered_mime_types
from everest.mime import get_registered_representer_names
from everest.mime import register_mime_type
from everest.testing import Pep8CompliantTestCase
from zope.interface import provider # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['MimeTestCase',
           ]


class MimeTestCase(Pep8CompliantTestCase):
    def test_registry(self):
        self.assert_raises(ValueError, register_mime_type,
                           MimeNotImplementingIMime)
        self.assert_raises(ValueError, register_mime_type,
                           MimeWithDuplicateTypeString)
        self.assert_raises(ValueError, register_mime_type,
                           MimeWithDuplicateNameString)
        self.assert_raises(ValueError, register_mime_type,
                           MimeWithDuplicateFileExtensionString)
        self.assert_true(XmlMime.mime_type_string
                                        in get_registered_mime_strings())
        self.assert_true(XmlMime.representer_name
                                        in get_registered_representer_names())
        self.assert_true(XmlMime in get_registered_mime_types())
        self.assert_true(
                get_registered_mime_type_for_string(XmlMime.mime_type_string)
                is XmlMime)
        self.assert_true(
                get_registered_mime_type_for_name(XmlMime.representer_name)
                is XmlMime)
        self.assert_equal(
                get_registered_mime_type_for_extension(XmlMime.file_extension),
                XmlMime)


class MimeNotImplementingIMime(object):
    pass


@provider(IMime)
class MimeWithDuplicateTypeString(object):
    mime_type_string = 'application/xml'
    representer_name = 'myxml'
    file_extension = 'xmlish'


@provider(IMime)
class MimeWithDuplicateNameString(object):
    mime_type_string = 'application/xmlish'
    representer_name = 'xml'
    file_extension = '.xmlish'


@provider(IMime)
class MimeWithDuplicateFileExtensionString(object):
    mime_type_string = 'application/xmlish'
    representer_name = 'myxml'
    file_extension = '.xml'



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 21, 2013.
"""
from everest.testing import Pep8CompliantTestCase
from everest.traversalpath import TraversalPath

__docformat__ = 'reStructuredText en'
__all__ = ['TraversalPathTestCase',
           ]


class TraversalPathTestCase(Pep8CompliantTestCase):
    def test_basics(self):
        tp = TraversalPath()
        self.assert_equal(len(tp), 0)
        self.assert_is_none(tp.parent)
        self.assert_is_none(tp.relation_operation)
        parent = 'proxy'
        attr = 'attribute'
        key = ('source', 'target')
        rel_op = 'relation_operation'
        tp.push(parent, key, attr, rel_op)
        self.assert_equal(len(tp), 1)
        self.assert_true(key in tp)
        self.assert_equal(tp.parent, parent)
        self.assert_equal(tp.relation_operation, rel_op)
        tp.pop()
        self.assert_equal(len(tp), 0)
        self.assert_false(key in tp)



"""
Utilities for the rdb backend.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.system import UserMessage
from everest.repositories.utils import GlobalObjectManager
from inspect import isdatadescriptor
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import func as sa_func
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import clear_mappers as sa_clear_mappers
from sqlalchemy.orm import mapper as sa_mapper
from sqlalchemy.orm.interfaces import MANYTOMANY
from sqlalchemy.orm.interfaces import MANYTOONE
from sqlalchemy.orm.interfaces import ONETOMANY
from sqlalchemy.orm.mapper import _mapper_registry
from sqlalchemy.sql.expression import cast
from threading import Lock

__docformat__ = 'reStructuredText en'
__all__ = ['OrmAttributeInspector',
           'as_slug_expression',
           'clear_mappers',
           'empty_metadata',
           'get_metadata',
           'hybrid_descriptor',
           'is_metadata_initialized',
           'map_system_entities',
           'mapper',
           'reset_metadata',
           'set_metadata',
           'synonym',
           ]


class _MetaDataManager(GlobalObjectManager):
    _globs = {}
    _lock = Lock()

    @classmethod
    def reset(cls):
        # This removes all attribute instrumentation from the entity classes.
        clear_mappers()
        # This is *very* important - the ORM attribute inspector caches
        # attributes which have become invalidated by the clearing of the
        # mappers.
        OrmAttributeInspector.reset()
        for md in cls._globs.values():
            md.clear()
        super(_MetaDataManager, cls).reset()

get_metadata = _MetaDataManager.get
set_metadata = _MetaDataManager.set
is_metadata_initialized = _MetaDataManager.is_initialized
reset_metadata = _MetaDataManager.reset


def clear_mappers():
    """
    Clears all mappers set up by SA and also clears all custom "id" and
    "slug" attributes inserted by the :func:`mapper` function in this module.

    This should only ever be needed in a testing context.
    """
    # Remove our hybrid property constructs.
    for mpr, is_primary in _mapper_registry.items():
        if is_primary:
            for attr_name in ('id', 'slug'):
                try:
                    attr = object.__getattribute__(mpr.class_, attr_name)
                    if isinstance(attr, hybrid_property):
                        if attr_name == 'id':
                            delattr(mpr.class_, attr_name)
                        else:
                            setattr(mpr.class_, attr_name, attr.descriptor)
                except AttributeError:
                    pass
    sa_clear_mappers()


def as_slug_expression(attr):
    """
    Converts the given instrumented string attribute into an SQL expression
    that can be used as a slug.

    Slugs are identifiers for members in a collection that can be used in an
    URL. We create slug columns by replacing non-URL characters with dashes
    and lower casing the result. We need this at the ORM level so that we can
    use the slug in a query expression.
    """
    slug_expr = sa_func.replace(attr, ' ', '-')
    slug_expr = sa_func.replace(slug_expr, '_', '-')
    slug_expr = sa_func.lower(slug_expr)
    return slug_expr


class hybrid_descriptor(hybrid_property):
    """
    Helper class wrapping a data descriptor into a hybrid property.
    """
    def __init__(self, descriptor, expr=None):
        self.__descriptor = descriptor
        hybrid_property.__init__(self, descriptor.fget,
                                 fset=descriptor.fset, fdel=descriptor.fdel,
                                 expr=expr)

    @property
    def descriptor(self):
        return self.__descriptor


def mapper(class_, local_table=None, id_attribute='id', slug_expression=None,
           *args, **kwargs):
    """
    Convenience wrapper around the SA mapper which will set up the hybrid
    "id" and "slug" attributes required by everest after calling the SA
    mapper.

    If you (e.g., for testing purposes) want to clear mappers created with
    this function, use the :func:`clear_mappers` function in this module.

    :param str id_attribute: the name of the column in the table to use as
      ID column (will be aliased to a new "id" attribute in the mapped class)
    :param slug_expression: function to generate a slug SQL expression given
      the mapped class as argument.
    """
    mpr = sa_mapper(class_, local_table=local_table, *args, **kwargs)
    # Set up the ID attribute as a hybrid property, if necessary.
    if id_attribute != 'id':
        # Make sure we are not overwriting an already mapped or customized
        # 'id' attribute.
        if 'id' in mpr.columns:
            mpr.dispose()
            raise ValueError('Attempting to overwrite the mapped "id" '
                             'attribute.')
        elif isdatadescriptor(getattr(class_, 'id', None)):
            mpr.dispose()
            raise ValueError('Attempting to overwrite the custom data '
                             'descriptor defined for the "id" attribute.')
        class_.id = synonym(id_attribute)
    # If this is a polymorphic class, a base class may already have a
    # hybrid descriptor set as slug attribute.
    slug_descr = None
    for base_cls in class_.__mro__:
        try:
            slug_descr = object.__getattribute__(base_cls, 'slug')
        except AttributeError:
            pass
        else:
            break
    if isinstance(slug_descr, hybrid_descriptor):
        if not slug_expression is None:
            raise ValueError('Attempting to overwrite the expression for '
                             'an inherited slug hybrid descriptor.')
        hyb_descr = slug_descr
    else:
        # Set up the slug attribute as a hybrid property.
        if slug_expression is None:
            cls_expr = lambda cls: cast(getattr(cls, 'id'), String)
        else:
            cls_expr = slug_expression
        hyb_descr = hybrid_descriptor(slug_descr, expr=cls_expr)
    class_.slug = hyb_descr
    return mpr


def synonym(name):
    """
    Utility function mimicking the behavior of the old SA synonym function
    with the new hybrid property semantics.
    """
    return hybrid_property(lambda inst: getattr(inst, name),
                           lambda inst, value: setattr(inst, name, value),
                           expr=lambda cls: getattr(cls, name))


def map_system_entities(engine, metadata, reset):
    """
    Maps all system entities.
    """
    # Map the user message system entity.
    msg_tbl = Table('_user_messages', metadata,
                    Column('guid', String, nullable=False, primary_key=True),
                    Column('text', String, nullable=False),
                    Column('time_stamp', DateTime(timezone=True),
                           nullable=False, default=sa_func.now()),
                    )
    mapper(UserMessage, msg_tbl, id_attribute='guid')
    if reset:
        metadata.drop_all(bind=engine, tables=[msg_tbl])
    metadata.create_all(bind=engine, tables=[msg_tbl])


def empty_metadata(engine):
    """
    The default metadata factory.
    """
    metadata = MetaData()
    metadata.create_all(bind=engine)
    return metadata


class OrmAttributeInspector(object):
    """
    Helper class inspecting class attributes mapped by the ORM.
    """
    __cache = {}

    @staticmethod
    def reset():
        """
        This clears the attribute cache this inspector maintains.

        Only needed in a testing context.
        """
        OrmAttributeInspector.__cache.clear()

    @staticmethod
    def inspect(orm_class, attribute_name):
        """
        :param attribute_name: name of the mapped attribute to inspect.
        :returns: list of 2-tuples containing information about the inspected
          attribute (first element: mapped entity attribute kind; second
          attribute: mapped entity attribute)
        """
        key = (orm_class, attribute_name)
        elems = OrmAttributeInspector.__cache.get(key)
        if elems is None:
            elems = OrmAttributeInspector.__inspect(key)
            OrmAttributeInspector.__cache[key] = elems
        return elems

    @staticmethod
    def __inspect(key):
        orm_class, attribute_name = key
        elems = []
        entity_type = orm_class
        ent_attr_tokens = attribute_name.split('.')
        count = len(ent_attr_tokens)
        for idx, ent_attr_token in enumerate(ent_attr_tokens):
            entity_attr = getattr(entity_type, ent_attr_token)
            kind, attr_type = OrmAttributeInspector.__classify(entity_attr)
            if idx == count - 1:
                pass
                # We are at the last name token - this must be a TERMINAL
                # or an ENTITY.
#                if kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
#                    raise ValueError('Invalid attribute name "%s": the '
#                                     'last element (%s) references an '
#                                     'aggregate attribute.'
#                                     % (attribute_name, ent_attr_token))
            else:
                if kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                    # We should not get here - the last attribute was a
                    # terminal.
                    raise ValueError('Invalid attribute name "%s": the '
                                     'element "%s" references a terminal '
                                     'attribute.'
                                     % (attribute_name, ent_attr_token))
                entity_type = attr_type
            elems.append((kind, entity_attr))
        return elems

    @staticmethod
    def __classify(attr):
        # Looks up the entity attribute kind and target type for the given
        # entity attribute.
        # We look for an attribute "property" to identify mapped attributes
        # (instrumented attributes and attribute proxies).
        if not hasattr(attr, 'property'):
            raise ValueError('Attribute "%s" is not mapped.' % attr)
        # We detect terminals by the absence of an "argument" attribute of
        # the attribute's property.
        if not hasattr(attr.property, 'argument'):
            kind = RESOURCE_ATTRIBUTE_KINDS.TERMINAL
            target_type = None
        else: # We have a relationship.
            target_type = attr.property.argument
            if attr.property.direction in (ONETOMANY, MANYTOMANY):
                if not attr.property.uselist:
                    # 1:1
                    kind = RESOURCE_ATTRIBUTE_KINDS.MEMBER
                else:
                    # 1:n or n:m
                    kind = RESOURCE_ATTRIBUTE_KINDS.COLLECTION
            elif attr.property.direction == MANYTOONE:
                kind = RESOURCE_ATTRIBUTE_KINDS.MEMBER
            else:
                raise ValueError('Unsupported relationship direction "%s".' # pragma: no cover
                                 % attr.property.direction)
        return kind, target_type



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 13, 2013.
"""
from everest.entities.attributes import get_domain_class_attribute
from everest.entities.relationship import DomainRelationship
from everest.querying.specifications import ValueContainedFilterSpecification
from everest.querying.specifications import ValueContainsFilterSpecification
from everest.querying.specifications import ValueEqualToFilterSpecification
from everest.resources.attributes import get_resource_class_attribute
from everest.resources.relationship import ResourceRelationship
from everest.resources.utils import get_root_collection
from everest.testing import EntityTestCase
from everest.testing import ResourceTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityParent
from mock import patch

__docformat__ = 'reStructuredText en'
__all__ = ['DomainRelationshipTestCase',
           ]


class DomainRelationshipTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_entity_attr(self):
        attr = get_domain_class_attribute(IMyEntity, 'parent')
        my_parent = MyEntityParent(id=0)
        self._run_test(attr, my_parent, 'child',
                       ValueEqualToFilterSpecification,
                       ValueEqualToFilterSpecification, 0)
        attr = get_domain_class_attribute(IMyEntityChild, 'parent')
        my_child = MyEntityChild(id=1)
        self._run_test(attr, my_child, 'children',
                       ValueContainsFilterSpecification,
                       ValueEqualToFilterSpecification, 1)

    def test_aggregate_attr(self):
        attr = get_domain_class_attribute(IMyEntity, 'children')
        my_child = MyEntityChild(id=0)
        self._run_test(attr, my_child, 'parent',
                       ValueEqualToFilterSpecification,
                       ValueContainedFilterSpecification, [0])

    def test_string_representation(self):
        attr = get_domain_class_attribute(IMyEntity, 'children')
        my_child = MyEntityChild(id=0)
        rel = attr.make_relationship(my_child)
        self.assert_true('<->' in str(rel))

    def _run_test(self, attr, relatee, backref_attr_name,
                  backref_relatee_spec_class,
                  ref_relatee_spec_class, relatee_spec_value):
        my_ent = MyEntity(id=0)
        rel = attr.make_relationship(my_ent)
        # With a backref, the spec references the relator.
        self.assert_true(isinstance(rel.specification,
                                    backref_relatee_spec_class))
        self.assert_equal(rel.specification.attr_name, backref_attr_name)
        self.assert_equal(rel.specification.attr_value, my_ent)
        # Without a backref, the spec references the relatee.
        with patch.object(attr.__class__, 'entity_backref', None):
            self.assert_true(isinstance(rel.specification,
                                        ValueEqualToFilterSpecification))
            self.assert_equal(rel.specification.attr_name, 'id')
            # Without a relatee, we get a non-matching spec.
            self.assert_equal(rel.specification.attr_value, -1)
            # When a relatee, its ID becomes available in the spec.
            rel.add(relatee)
            self.assert_true(isinstance(rel.specification,
                                        ref_relatee_spec_class))
            self.assert_equal(rel.specification.attr_value,
                              relatee_spec_value)


class ResourceRelationshipTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_make_relationship(self):
        ent = MyEntity(id=0)
        coll = get_root_collection(IMyEntity)
        mb = coll.create_member(ent)
        attr = get_resource_class_attribute(IMyEntity, 'parent')
        rel = attr.make_relationship(mb)
        self.assert_true(isinstance(rel, ResourceRelationship))
        parent = MyEntityParent(id=0)
        parent_coll = get_root_collection(IMyEntityParent)
        parent_mb = parent_coll.create_member(parent)
        self.assert_true(ent.parent is None)
        rel.add(parent_mb)
        self.assert_true(ent.parent is parent)
        rel.remove(parent_mb)
        self.assert_true(ent.parent is None)
        rel = attr.make_relationship(ent)
        self.assert_true(isinstance(rel, DomainRelationship))
        self.assert_raises(ValueError, attr.make_relationship, None)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 5, 2012.
"""
from everest.mime import CsvMime
from everest.renderers import RendererFactory
from everest.renderers import ResourceRenderer
from everest.testing import ResourceTestCase
from everest.tests.complete_app.testing import create_collection
from pyramid.compat import binary_type

__docformat__ = 'reStructuredText en'
__all__ = ['RendererTestCase',
           ]


class RendererTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'
    class _Info(object):
        def __init__(self, name):
            self.name = name
    def test_renderer_factory(self):
        def _test(rc, rnd_name, exp_str):
            inf = self._Info(rnd_name)
            fac = RendererFactory(inf)
            rpr_str = fac(dict(), dict(request=self._request, context=rc))
            self.assert_true(isinstance(rpr_str, binary_type))
            self.assert_not_equal(rpr_str.find(exp_str), -1)
        resource = create_collection()
        _test(resource, 'csv', b'"id","parent.id"')
        _test(resource, 'atom', b'www.w3.org/2005/Atom')
        self.assert_raises(ValueError, _test, resource, 'foo', '')
        self.assert_raises(ValueError, _test, None, 'csv', '"id","parent"')

    def test_nonvalidating_renderer(self):
        class FooRenderer(ResourceRenderer):
            def __init__(self):
                ResourceRenderer.__init__(self, CsvMime)

            def _validate(self, resource): # pylint: disable=W0613
                return False

            def _prepare_response(self, system):
                pass
        rc = create_collection()
        rnd = FooRenderer()
        self.assert_raises(ValueError, rnd,
                           dict(), dict(request=self._request, context=rc))



"""
Interfaces for classes related to the querying subsystem.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 2, 2011.
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['IFilterSpecificationFactory',
           'IFilterSpecificationVisitor',
           'IOrderSpecificationFactory',
           'IOrderSpecificationVisitor',
           'ISpecification',
           ]


# begin interfaces pylint: disable=E0213,W0232,E0211

class ISpecification(Interface):
    """
    Specification interface.
    """
    operator = Attribute('The operator for this specification. Subclass of '
                         ':class:`everest.querying.operators.Operator`.')

    def accept(visitor):
        """
        Accept the given visitor into this specification.

        This triggers visits of this specification and all other dependent
        specifications which in turn dispatch appropriate visiting operations.

        :param visitor: a visitor that packages related operations
        :type visitor: object implementing :class:`ISpecificationVisitor`
        """


class IFilterSpecificationFactory(Interface):
    """
    Filter specification factory interface.
    """
    def create_equal_to(attr_name, attr_value):
        "Create an equal-to filter specification."

    def create_starts_with(attr_name, attr_value):
        "Create a starts-with filter specification."

    def create_ends_with(attr_name, attr_value):
        "Create an ends-with filter specification."

    def create_contains(attr_name, attr_value):
        "Create a contains filter specification."

    def create_contained(attr_name, attr_value):
        "Create a contained filter specification."

    def create_greater_than_or_equal_to(attr_name, attr_value):
        "Create a greater-than-or-equal-to filter specification."

    def create_greater_than(attr_name, attr_value):
        "Create a greater-than filter specification."

    def create_less_than_or_equal_to(attr_name, attr_value):
        "Create a less-than-or-equal-to filter specification."

    def create_less_than(attr_name, attr_value):
        "Create a less-than filter specification."

    def create_in_range(attr_name, from_value, to_value):
        "Create an in-range filter specification."
        raise NotImplementedError('Abstract method')

    def create_conjunction(left_spec, right_spec):
        "Create a conjunction of two filter specifications."

    def create_disjunction(left_spec, right_spec):
        "Create a disjunction of two filter specifications."

    def create_negation(wrapped):
        "Create a negation of a filter specification."


class IOrderSpecificationFactory(Interface):
    """
    Order specification factory interface.
    """
    def create_ascending(attr_name):
        "Create an ascending order specification."

    def create_descending(attr_name):
        "Create a descending order specification."



class ISpecificationVisitor(Interface):
    """
    Interface for specification visitors.

    The various visiting methods dispatch to the appropriate visiting
    operations depending on the passed specification's operator.
    """
    expression = Attribute('The expression the visitor built.')

    def visit_nullary(spec):
        """
        Visits the given specification with a dispatched visiting operation.
        """

    def visit_unary(spec):
        """
        Visits the given specification, passing the expression obtained from
        processing the last specification as an argument to the visiting
        operation.
        """

    def visit_binary(spec):
        """
        Visits the given specification, passing the expressions obtained from
        processing the last two specifications as an arguments to the visiting
        operation.
        """


class IFilterSpecificationVisitor(ISpecificationVisitor):
    """
    Interface for filter specification visitors.
    """
    def filter_query(query):
        """
        Returns the given query filtered by the expression built by this
        visitor.

        Needs to be called after :method:`accept` has been run.
        """

    def _conjunction_op(spec, *expressions):
        """
        Visiting operation for conjunction specifications.
        """

    def _disjunction_op(spec, *expressions):
        """
        Visiting operation for disjunction specifications.
        """

    def _negation_op(spec, expression):
        """
        Visiting operation for negation specifications.
        """

    def _starts_with_op(spec):
        """
        Visiting operation for value starts with specifications.
        """

    def _ends_with_op(spec):
        """
        Visiting operation for value ends with specifications.
        """

    def _contains_op(spec):
        """
        Visiting operation for value contains specifications.
        """

    def _contained_op(spec):
        """
        Visiting operation for value contained specifications.
        """

    def _equal_to_op(spec):
        """
        Visiting operation for value equal to specifications.
        """

    def _less_than_op(spec):
        """
        Visiting operation for value less than specifications.
        """

    def _less_than_or_equal_to_op(spec):
        """
        Visiting operation for value less than or equal to specifications.
        """

    def _greater_than_op(spec):
        """
        Visiting operation for value greater than specifications.
        """

    def _greater_than_or_equal_to_op(spec):
        """
        Visiting operation for value greater than or equal to specifications.
        """

    def _in_range_op(spec):
        """
        Visiting operation for value in range specifications.
        """


class IOrderSpecificationVisitor(ISpecificationVisitor):
    """
    Interface for order specification visitors that generate a query
    expression.
    """
    def order_query(query):
        """
        Returns the given query ordered by the expression built by this
        visitor.

        Needs to be called after :method:`accept` has been run.
        """

    def _conjunction_op(spec, *expressions):
        """
        Visiting operation for conjunction specifications.
        """

    def _asc_op(attr_name):
        """
        Visiting operation for ascending order specifications.
        """

    def _desc_op(attr_name):
        """
        Visiting operation for descending order specifications.
        """


class IQuery(Interface):
    """
    Interface for everest queries.
    """
    _entity_class = Attribute('The entity class the query targets.')

    def __iter__():
        """
        Returns an iterator over all entities in this query after applying
        filtering, ordering, and slicing settings.
        """

    def count():
        """
        Returns the count of the entities in this query.

        :note: This does not take slicing into account.
        """

    def all():
        """
        Returns a list of all entities in this query after applying
        filtering, ordering, and slicing settings.
        """

    def one():
        """
        Returns exactly one result from this query.

        :raises NoResultsException: if no results were found.
        :raises MultipleResultsException: if more than one result was found.
        """

    def filter(filter_expression):
        """
        Sets the filter expression for this query. Generative (returns a
        clone).

        :note: If the query already has a filter expression, the returned
            query will use the conjunction of both expressions.
        """

    def filter_by(**kw):
        """
        Generates an equal-to filter expression (or a conjunction from
        multiple equal-to filter expressions) and calls :method:`filter`
        with it.

        :param dict kw: name, value pairs.
        """

    def order(order_expression):
        """
        Sets the order expression for this query. Generative (returns a
        clone).

        :note: If the query already has an order expression, the returned
            query will use the conjunction of both expressions.
        """

    def order_by(*args):
        """
        Generates an ordering expression (or a conjunction from multiple
        ordering expressions) and calls :method:`order` with it.

        :param tuple args: tuple containing (name, ascending) operator pairs.
        """

    def slice(start, stop):
        """
        Sets the slice key for this query. Generative (returns a clone).
        """

# end interfaces pylint: enable=E0213,W0232,E0211



"""
Interfaces for views.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 22, 2010.
"""
from zope.interface import Interface, Attribute # pylint: disable=E0611,F0401

__docformat__ = "reStructuredText en"
__all__ = ['IResourceView',
           ]


# no __init__, no self. pylint: disable=W0232,E0211

class IResourceView(Interface):
    """
    Interface for resource views.
    """

    context = Attribute("The request's context")
    request = Attribute("The context from the object graph")

    def __call__():
        """
        """
# pylint: disable=W0232,E0211



"""
MIME (content) types.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Aug 17, 2011.
"""
from everest.interfaces import IAtomEntryMime
from everest.interfaces import IAtomFeedMime
from everest.interfaces import IAtomMime
from everest.interfaces import IAtomRequest
from everest.interfaces import IAtomServiceMime
from everest.interfaces import ICsvMime
from everest.interfaces import ICsvRequest
from everest.interfaces import IHtmlMime
from everest.interfaces import IHtmlRequest
from everest.interfaces import IJsonMime
from everest.interfaces import IJsonRequest
from everest.interfaces import IMime
from everest.interfaces import ITextPlainMime
from everest.interfaces import IXlsMime
from everest.interfaces import IXlsRequest
from everest.interfaces import IXmlMime
from everest.interfaces import IXmlRequest
from everest.interfaces import IZipMime
from everest.interfaces import IZipRequest
from everest.utils import BidirectionalLookup
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface import provider # pylint: disable=E0611,F0401

__docformat__ = "reStructuredText en"
__all__ = ['ATOM_FEED_MIME',
           'ATOM_ENTRY_MIME',
           'ATOM_MIME',
           'AtomEntryMime',
           'AtomFeedMime',
           'AtomMime',
           'AtomServiceMime',
           'CSV_MIME',
           'CsvMime',
           'HTML_MIME',
           'HtmlMime',
           'JSON_MIME',
           'JsonMime',
           'MIME_REQUEST',
           'TEXT_PLAIN_MIME',
           'TextPlainMime',
           'XLS_MIME',
           'XML_MIME',
           'XlsMime',
           'XmlMime',
           'ZIP_MIME',
           'ZipMime',
           'get_registered_mime_strings',
           'get_registered_mime_type_for_extension',
           'get_registered_mime_type_for_name',
           'get_registered_mime_type_for_string',
           'get_registered_mime_types',
           'get_registered_representer_names',
           'register_mime_type',
           ]


class MimeTypeRegistry(object):
    """
    Simple registry for MIME content types.
    """
    __type_string_map = {}
    __rpr_name_map = {}
    __file_extension_map = BidirectionalLookup()

    @classmethod
    def register(cls, mime_type):
        if not [ifc for ifc in provided_by(mime_type)
                if issubclass(ifc, IMime)]:
            raise ValueError('MIME type to register must implement the '
                             'IMime interface.')
        if mime_type.mime_type_string in cls.__type_string_map:
            raise ValueError('Duplicate MIME string detected.')
        if mime_type.representer_name in cls.__rpr_name_map:
            raise ValueError('Duplicate MIME name detected.')
        if mime_type.file_extension in cls.__file_extension_map:
            raise ValueError('Duplicate file extension detected.')
        cls.__type_string_map[mime_type.mime_type_string] = mime_type
        cls.__rpr_name_map[mime_type.representer_name] = mime_type
        cls.__file_extension_map[mime_type.file_extension] = mime_type

    @classmethod
    def get_types(cls):
        return cls.__type_string_map.values()

    @classmethod
    def get_strings(cls):
        return cls.__type_string_map.keys()

    @classmethod
    def get_names(cls):
        return cls.__rpr_name_map.keys()

    @classmethod
    def get_type_for_string(cls, mime_type_string):
        return cls.__type_string_map[mime_type_string]

    @classmethod
    def get_type_for_name(cls, representer_name):
        return cls.__rpr_name_map[representer_name]

    @classmethod
    def get_type_for_extension(cls, file_extension):
        return cls.__file_extension_map[file_extension]


register_mime_type = MimeTypeRegistry.register
get_registered_mime_strings = MimeTypeRegistry.get_strings
get_registered_representer_names = MimeTypeRegistry.get_names
get_registered_mime_types = MimeTypeRegistry.get_types
get_registered_mime_type_for_string = MimeTypeRegistry.get_type_for_string
get_registered_mime_type_for_name = MimeTypeRegistry.get_type_for_name
get_registered_mime_type_for_extension = \
                                    MimeTypeRegistry.get_type_for_extension


@provider(IJsonMime)
class JsonMime(object):
    mime_type_string = 'application/json'
    representer_name = 'json'
    file_extension = '.json'

JSON_MIME = JsonMime.mime_type_string

register_mime_type(JsonMime)


@provider(IAtomMime)
class AtomMime(object):
    mime_type_string = 'application/atom+xml'
    representer_name = 'atom'
    file_extension = '.atom'

ATOM_MIME = AtomMime.mime_type_string

register_mime_type(AtomMime)


class AtomFeedMime(AtomMime):
    provider(IAtomFeedMime)
    mime_type_string = 'application/atom+xml;type=feed'

ATOM_FEED_MIME = AtomFeedMime.mime_type_string


class AtomEntryMime(AtomMime):
    provider(IAtomEntryMime)
    mime_type_string = 'application/atom+xml;type=entry'

ATOM_ENTRY_MIME = AtomFeedMime.mime_type_string


class AtomServiceMime(AtomMime):
    provider(IAtomServiceMime)
    mime_type_string = 'application/atomsvc+xml'

ATOM_SERVICE_MIME = AtomFeedMime.mime_type_string


@provider(IXmlMime)
class XmlMime(object):
    mime_type_string = 'application/xml'
    representer_name = 'xml'
    file_extension = '.xml'

XML_MIME = XmlMime.mime_type_string

register_mime_type(XmlMime)


@provider(ICsvMime)
class CsvMime(object):
    mime_type_string = 'application/csv'
    representer_name = 'csv'
    file_extension = '.csv'

CSV_MIME = CsvMime.mime_type_string

register_mime_type(CsvMime)


@provider(IHtmlMime)
class HtmlMime(object):
    mime_type_string = 'text/html'
    file_extension = '.html'

HTML_MIME = HtmlMime.mime_type_string


@provider(ITextPlainMime)
class TextPlainMime(object):
    mime_type_string = 'text/plain'
    file_extension = '.txt'

TEXT_PLAIN_MIME = TextPlainMime.mime_type_string


@provider(IXlsMime)
class XlsMime(object):
    mime_type_string = 'application/vnd.xls'
    file_extension = '.xls'

XLS_MIME = XlsMime.mime_type_string


@provider(IZipMime)
class ZipMime(object):
    mime_type_string = 'application/zip'
    file_extension = '.zip'

ZIP_MIME = ZipMime.mime_type_string


MIME_REQUEST = {JSON_MIME : IJsonRequest,
                ATOM_MIME : IAtomRequest,
                ATOM_FEED_MIME : IAtomRequest,
                ATOM_ENTRY_MIME : IAtomRequest,
                ATOM_SERVICE_MIME : IAtomRequest,
                XML_MIME : IXmlRequest,
                CSV_MIME : ICsvRequest,
                HTML_MIME : IHtmlRequest,
                XLS_MIME : IXlsRequest,
                ZIP_MIME : IZipRequest,
                }



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.

Package initialization file.
"""
from .aggregate import MemoryAggregate as Aggregate
from .querying import ObjectFilterSpecificationVisitor
from .querying import ObjectOrderSpecificationVisitor
from .repository import MemoryRepository as Repository
from .session import MemorySession as Session



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 11, 2013.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.attributes import get_domain_class_collection_attribute_iterator
from everest.entities.attributes import get_domain_class_attribute
from everest.entities.attributes import get_domain_class_attribute_names
from everest.entities.attributes import get_domain_class_attributes
from everest.entities.attributes import get_domain_class_relationship_attribute_iterator
from everest.entities.attributes import get_domain_class_member_attribute_iterator
from everest.entities.attributes import get_domain_class_terminal_attribute_iterator
from everest.entities.attributes import is_domain_class_collection_attribute
from everest.entities.attributes import is_domain_class_domain_attribute
from everest.entities.attributes import is_domain_class_member_attribute
from everest.entities.attributes import is_domain_class_terminal_attribute
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.testing import EntityTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.interfaces import IMyEntity

__docformat__ = 'reStructuredText en'
__all__ = ['EntityAttributesTestCase',
           ]


class EntityAttributesTestCase(RdbTestCaseMixin, EntityTestCase):
    package_name = 'everest.tests.complete_app'

    def _test_accessors(self, obj):
        names = list(get_domain_class_attribute_names(obj))
        self.assert_equal(list(names),
                          ['id', 'parent', 'children', 'text', 'text_ent',
                           'number', 'date_time', 'parent.text_ent'])
        self.assert_equal(list(get_domain_class_attributes(obj).keys()),
                          names)
        term_attr_name = 'number'
        term_attr = get_domain_class_attribute(obj, term_attr_name)
        self.assert_equal(term_attr.entity_attr, term_attr_name)
        self.assert_true(
                    is_domain_class_terminal_attribute(obj, term_attr_name))
        ent_attr_name = 'parent'
        ent_attr = get_domain_class_attribute(obj, ent_attr_name)
        self.assert_equal(ent_attr.entity_attr, ent_attr_name)
        self.assert_true(
                    is_domain_class_member_attribute(obj, ent_attr_name))
        self.assert_true(
                    is_domain_class_domain_attribute(obj, ent_attr_name))
        agg_attr_name = 'children'
        agg_attr = get_domain_class_attribute(obj, agg_attr_name)
        self.assert_equal(agg_attr.entity_attr, agg_attr_name)
        self.assert_true(
                    is_domain_class_collection_attribute(obj, agg_attr_name))
        self.assert_true(
                    is_domain_class_domain_attribute(obj, agg_attr_name))
        for attr in get_domain_class_terminal_attribute_iterator(obj):
            self.assert_equal(attr.kind, RESOURCE_ATTRIBUTE_KINDS.TERMINAL)
        for attr in get_domain_class_member_attribute_iterator(obj):
            self.assert_equal(attr.kind, RESOURCE_ATTRIBUTE_KINDS.MEMBER)
        for attr in get_domain_class_collection_attribute_iterator(obj):
            self.assert_equal(attr.kind, RESOURCE_ATTRIBUTE_KINDS.COLLECTION)
        for attr in get_domain_class_relationship_attribute_iterator(obj):
            self.assert_true(attr.kind in (RESOURCE_ATTRIBUTE_KINDS.MEMBER,
                                           RESOURCE_ATTRIBUTE_KINDS.COLLECTION))

    def test_accessors_with_interface(self):
        self._test_accessors(IMyEntity)

    def test_accessors_with_entity_class(self):
        self._test_accessors(MyEntity)



"""
In-memory session.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 8, 2013.
"""
from collections import MutableSequence
from collections import MutableSet
from everest.constants import RELATION_OPERATIONS
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.traversal import AruVisitor
from everest.entities.utils import get_entity_class
from everest.repositories.base import AutocommittingSessionMixin
from everest.repositories.base import Session
from everest.repositories.base import SessionFactory
from everest.repositories.memory.cache import EntityCache
from everest.repositories.memory.querying import MemoryRepositoryQuery
from everest.repositories.state import EntityState
from everest.repositories.uow import UnitOfWork
from everest.traversal import SourceTargetDataTreeTraverser
from pyramid.compat import iteritems_
from threading import local
from transaction.interfaces import IDataManager
from zope.interface import implementer # pylint: disable=E0611,F0401
import transaction

__docformat__ = 'reStructuredText en'
__all__ = ['DataManager',
           'MemoryAutocommittingSession',
           'MemorySession',
           'MemorySessionFactory',
           ]


class MemorySession(Session):
    """
    Session object.

    The session
     * Holds a Unit Of Work;
     * Serves as identity and slug map;
     * Performs synchronized commit on repository;
     * Sets up data manager to hook into transaction.
    """
    IS_MANAGING_BACKREFERENCES = True

    def __init__(self, repository, query_class=None, clone_on_load=True):
        self.__repository = repository
        self.__unit_of_work = UnitOfWork()
        self.__cache_map = {}
        if query_class is None:
            query_class = MemoryRepositoryQuery
        self.__query_class = query_class
        self.__clone_on_load = clone_on_load
        self.__needs_flushing = False
        self.__is_flushing = False

    def get_by_id(self, entity_class, entity_id):
        if self.__needs_flushing:
            self.flush()
        cache = self.__get_cache(entity_class)
        return cache.get_by_id(entity_id)

    def get_by_slug(self, entity_class, entity_slug):
        # When the entity is not found in the cache, it may have been added
        # with an undefined slug; we therefore attempt to look it up in the
        # list of pending NEW entities.
        if self.__needs_flushing:
            self.flush()
        cache = self.__get_cache(entity_class)
        ent = cache.get_by_slug(entity_slug)
#        if ent is None:
#            for new_ent in self.__unit_of_work.get_new(entity_class):
#                if new_ent.slug == entity_slug:
#                    ent = new_ent
#                    break
        return ent

    def add(self, entity_class, data):
        self.__traverse(entity_class, data, None, RELATION_OPERATIONS.ADD)

    def remove(self, entity_class, data):
        self.__traverse(entity_class, None, data, RELATION_OPERATIONS.REMOVE)

    def update(self, entity_class, data, target=None):
        return self.__traverse(entity_class, data, target,
                               RELATION_OPERATIONS.UPDATE)

    def query(self, entity_class):
        if self.__needs_flushing:
            self.flush()
        return self.__query_class(entity_class, self, self.__repository)

    def flush(self):
        if self.__needs_flushing and not self.__is_flushing:
            self.__is_flushing = True
            with self.__repository.lock:
                self.__repository.flush(self.__unit_of_work)
            self.__is_flushing = False
            for ent_cls in self.__cache_map.keys():
                # The flush may have auto-generated IDs for NEW entities,
                # so we rebuild the cache.
                cache = self.__get_cache(ent_cls)
                cache.rebuild(self.__unit_of_work.get_new(ent_cls))
        self.__needs_flushing = False

    def begin(self):
        self.__unit_of_work.reset()

    def commit(self):
        with self.__repository.lock:
            self.__repository.commit(self.__unit_of_work)
        self.__unit_of_work.reset()
        self.__cache_map.clear()

    def rollback(self):
        with self.__repository.lock:
            self.__repository.rollback(self.__unit_of_work)
        self.__unit_of_work.reset()
        self.__cache_map.clear()

    def reset(self):
        self.rollback()

    def load(self, entity_class, entity):
        """
        Load the given repository entity into the session and return a
        clone. If it was already loaded before, look up the loaded entity
        and return it.

        All entities referenced by the loaded entity will also be loaded
        (and cloned) recursively.

        :raises ValueError: When an attempt is made to load an entity that
          has no ID
        """
        if self.__needs_flushing:
            self.flush()
        if entity.id is None:
            raise ValueError('Can not load entity without an ID.')
        cache = self.__get_cache(entity_class)
        sess_ent = cache.get_by_id(entity.id)
        if sess_ent is None:
            if self.__clone_on_load:
                sess_ent = self.__clone(entity, cache)
            else: # Only needed by the nosql backend pragma: no cover
                cache.add(entity)
                sess_ent = entity
            self.__unit_of_work.register_clean(entity_class, sess_ent)
        return sess_ent

    @property
    def new(self):
        return self.__unit_of_work.get_new()

    @property
    def deleted(self):
        return self.__unit_of_work.get_deleted()

    def __contains__(self, entity):
        cache = self.__cache_map.get(type(entity))
        if not cache is None:
            found = entity in cache
        else:
            found = False
        return found

    def __traverse(self, entity_class, source_data, target_data, rel_op):
        agg = self.__repository.get_aggregate(entity_class)
        trv = SourceTargetDataTreeTraverser.make_traverser(source_data,
                                                           target_data,
                                                           rel_op,
                                                           accessor=agg)
        vst = AruVisitor(entity_class,
                         self.__add, self.__remove, self.__update)
        trv.run(vst)
        # Indicate that we need to flush the changes.
        self.__needs_flushing = True
        return vst.root

    def __add(self, entity_class, entity):
        cache = self.__get_cache(entity_class)
        # We allow adding the same entity multiple times.
        if not (not entity.id is None
                and cache.get_by_id(entity.id) is entity):
            if not self.__unit_of_work.is_marked_deleted(entity):
                self.__unit_of_work.register_new(entity_class, entity)
                if not entity.id is None and cache.has_id(entity.id):
                    raise ValueError('Duplicate entity ID "%s".' % entity.id)
                if not entity.slug is None and cache.has_slug(entity.slug):
                    raise ValueError('Duplicate entity slug "%s".'
                                     % entity.slug)
            else:
                if self.__unit_of_work.is_marked_pending(entity):
                    # The changes were not flushed yet; just mark as clean.
                    self.__unit_of_work.mark_clean(entity)
                else:
                    self.__unit_of_work.mark_new(entity)
                    self.__unit_of_work.mark_pending(entity)
            cache.add(entity)

    def __remove(self, entity_class, entity):
        if not self.__unit_of_work.is_registered(entity):
            if entity.id is None:
                raise ValueError('Can not remove un-registered entity '
                                 'without an ID')
            self.__unit_of_work.register_deleted(entity_class, entity)
        elif not self.__unit_of_work.is_marked_new(entity):
            self.__unit_of_work.mark_deleted(entity)
        else:
            if self.__unit_of_work.is_marked_pending(entity):
                # The changes were not flushed yet; just mark as clean.
                self.__unit_of_work.mark_clean(entity)
            else:
                self.__unit_of_work.mark_deleted(entity)
                self.__unit_of_work.mark_pending(entity)
        cache = self.__get_cache(entity_class)
        if entity in cache:
            cache.remove(entity)

    def __update(self, entity_class, source_data, target_entity): # pylint: disable=W0613
        EntityState.set_state_data(target_entity, source_data)
        if self.__unit_of_work.is_marked_persisted(target_entity):
            self.__unit_of_work.mark_pending(target_entity)

    def __get_cache(self, entity_class):
        cache = self.__cache_map.get(entity_class)
        if cache is None:
            cache = self.__cache_map[entity_class] = EntityCache()
        return cache

    def __clone(self, entity, cache):
        clone = object.__new__(entity.__class__)
        # We add the clone with its ID set to the cache *before* we load it
        # so that circular references will work.
        clone.id = entity.id
        cache.add(clone)
        state = EntityState.get_state_data(entity)
        id_attr = None
        for attr, value in iteritems_(state):
            if attr.entity_attr == 'id':
                id_attr = attr
                continue
            attr_type = attr.attr_type
            if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL \
               and not self.__repository.is_registered_resource(attr_type):
                # Prevent loading of entities from other repositories.
                # FIXME: Doing this here is inconsistent, since e.g. the RDB
                #        session does not perform this kind of check.
                continue
            elif attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER \
               and not value is None:
                ent_cls = get_entity_class(attr_type)
                new_value = self.load(ent_cls, value)
                state[attr] = new_value
            elif attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION \
                 and len(value) > 0:
                value_type = type(value)
                new_value = value_type.__new__(value_type)
                if issubclass(value_type, MutableSequence):
                    add_op = new_value.append
                elif issubclass(value_type, MutableSet):
                    add_op = new_value.add
                else:
                    raise ValueError('Do not know how to clone value of type '
                                     '%s for resource attribute %s.'
                                     % (type(new_value), attr))
                ent_cls = get_entity_class(attr_type)
                for child in value:
                    child_clone = self.load(ent_cls, child)
                    add_op(child_clone)
                state[attr] = new_value
        # We set the ID already above.
        if not id_attr is None:
            del state[id_attr]
        EntityState.set_state_data(clone, state)
        return clone


class MemoryAutocommittingSession(AutocommittingSessionMixin, MemorySession):
    """
    Autocommitting session in memory.
    """
    pass


class MemorySessionFactory(SessionFactory):
    """
    Factory for :class:`MemorySession` instances.

    The factory creates exactly one session per thread.
    """
    def __init__(self, repository, query_class=None, clone_on_load=True):
        SessionFactory.__init__(self, repository)
        sess_reg = local()
        self.__session_registry = sess_reg
        self.__query_class = query_class
        self.__clone_on_load = clone_on_load

    def __call__(self):
        session = getattr(self.__session_registry, 'session', None)
        if session is None:
            if not self._repository.autocommit:
                session = MemorySession(self._repository,
                                        query_class=self.__query_class,
                                        clone_on_load=self.__clone_on_load)
            else:
                session = MemoryAutocommittingSession(
                                        self._repository,
                                        query_class=self.__query_class,
                                        clone_on_load=self.__clone_on_load)
            self.__session_registry.session = session
            if self._repository.join_transaction is True:
                self.__session_registry.data_manager = DataManager(session)
        if self._repository.join_transaction is True:
            trx = transaction.get()
            dm = self.__session_registry.data_manager
            # We have a new transaction that we need to join.
            if not dm.transaction is trx:
                trx.join(dm)
                dm.transaction = trx
        return session


@implementer(IDataManager)
class DataManager(object):
    """
    Data manager to plug a :class:`MemorySession` into a Zope transaction.
    """
    # TODO: implement safepoints.

    def __init__(self, session):
        self.__session = session
        self.transaction = None

    def abort(self, trans): # pylint: disable=W0613
        self.__session.rollback()

    def tpc_begin(self, trans): # pylint: disable=W0613
        pass

    def commit(self, trans): # pylint: disable=W0613
        self.__session.commit()

    def tpc_vote(self, trans): # pylint: disable=W0613
        pass

    def tpc_finish(self, trans):
        pass

    def tpc_abort(self, trans): # pylint: disable=W0613
        self.__session.rollback()

    def sortKey(self):
        return "everest:%d" % id(self.__session)



"""
Delete member view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 24, 2011.
"""
from everest.utils import get_traceback
from everest.views.base import ResourceView
from pyramid.httpexceptions import HTTPOk

__docformat__ = 'reStructuredText en'
__all__ = ['DeleteMemberView',
           ]


class DeleteMemberView(ResourceView):
    """
    A View for processing DELETE requests

    The client sends a DELETE request to the URI of a Member Resource. If the
    deletion is successful, the server responds with a status code of 200.

    In a RESTful server DELETE does not always mean "delete a record from the
    database". See RESTful Web Services and REST in Practice books.

    See http://bitworking.org/projects/atom/rfc5023.html#delete-via-DELETE
    """

    def __call__(self):
        self._logger.debug('DELETE Request received on %s' % self.request.url)
        try:
            self.context.__parent__.remove(self.context)
        except Exception as err: # catch Exception pylint: disable=W0703
            response = self._handle_unknown_exception(str(err),
                                                      get_traceback())
        else:
            response = self.request.get_response(HTTPOk())
        return response




"""
Query specifications.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

The central idea of a Specification is to separate the statement of how to
match a candidate from the candidate object that it is matched against.

Read http://en.wikipedia.org/wiki/Specification_pattern for more info and
especially http://www.martinfowler.com/apsupp/spec.pdf

Created on Jul 5, 2011.
"""
import re

from pyramid.compat import string_types
from pyramid.threadlocal import get_current_registry

from everest.querying.interfaces import IFilterSpecificationFactory
from everest.querying.interfaces import IOrderSpecificationFactory
from everest.querying.interfaces import ISpecification
from everest.querying.operators import ASCENDING
from everest.querying.operators import CONJUNCTION
from everest.querying.operators import CONTAINED
from everest.querying.operators import CONTAINS
from everest.querying.operators import DESCENDING
from everest.querying.operators import DISJUNCTION
from everest.querying.operators import ENDS_WITH
from everest.querying.operators import EQUAL_TO
from everest.querying.operators import GREATER_OR_EQUALS
from everest.querying.operators import GREATER_THAN
from everest.querying.operators import IN_RANGE
from everest.querying.operators import LESS_OR_EQUALS
from everest.querying.operators import LESS_THAN
from everest.querying.operators import NEGATION
from everest.querying.operators import STARTS_WITH
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.utils import get_nested_attribute
from zope.interface import implementer # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['AscendingOrderSpecification',
           'CompositeFilterSpecification',
           'ConjunctionFilterSpecification',
           'ConjunctionOrderSpecification',
           'CriterionFilterSpecification',
           'DescendingOrderSpecification',
           'DisjunctionFilterSpecification',
           'FilterSpecification',
           'FilterSpecificationFactory',
           'LeafFilterSpecification',
           'NaturalOrderSpecification',
           'NegationFilterSpecification',
           'ObjectOrderSpecification',
           'OrderSpecification',
           'OrderSpecificationFactory',
           'Specification',
           'ValueContainedFilterSpecification',
           'ValueContainsFilterSpecification',
           'ValueEndsWithFilterSpecification',
           'ValueEqualToFilterSpecification',
           'ValueGreaterThanFilterSpecification',
           'ValueGreaterThanOrEqualToFilterSpecification',
           'ValueInRangeFilterSpecification',
           'ValueLessThanFilterSpecification',
           'ValueLessThanOrEqualToFilterSpecification',
           'ValueStartsWithFilterSpecification',
           'asc',
           'cnts',
           'cntd',
           'desc',
           'eq',
           'ends',
           'ge',
           'gt',
           'le',
           'lt',
           'order',
           'rng',
           'starts',
           ]


@implementer(ISpecification)
class Specification(object):
    """
    Abstract base classs for all specifications.
    """
    operator = None

    def __init__(self):
        if self.__class__ is Specification:
            raise NotImplementedError('Abstract class')

    def accept(self, visitor):
        raise NotImplementedError('Abstract method')


class FilterSpecification(Specification):
    """
    Abstract base class for all filter specifications.
    """
    def __init__(self):
        if self.__class__ is FilterSpecification:
            raise NotImplementedError('Abstract class')
        Specification.__init__(self)

    def is_satisfied_by(self, candidate):
        """
        Tells if the given candidate object matches this specification.

        :param candidate: the candidate object
        :type candidate: object
        :returns: True if the specification is met by the candidate.
        :rtype: bool
        """
        raise NotImplementedError('Abstract method')

    def __and__(self, other):
        return ConjunctionFilterSpecification(self, other)

    def __or__(self, other):
        return DisjunctionFilterSpecification(self, other)

    def __invert__(self):
        return NegationFilterSpecification(self)


class LeafFilterSpecification(FilterSpecification): # still abstract pylint: disable=W0223
    """
    Abstract base class for specifications that represent leaves in a
    specification tree.
    """
    def __init__(self):
        if self.__class__ is LeafFilterSpecification:
            raise NotImplementedError('Abstract class')
        FilterSpecification.__init__(self)

    def accept(self, visitor):
        visitor.visit_nullary(self)


class CriterionFilterSpecification(LeafFilterSpecification):
    """
    Abstract base class for specifications representing filter criteria.
    """
    def __init__(self, attr_name, attr_value):
        """
        Constructs a filter specification for a query criterion.

        :param operator: operator
        :type operator: :class:`everest.querying.operators.Operator`
        :param attr_name: the candidate's attribute name
        :type attr_name: str
        :param attr_value: the value that satisfies the specification
        :type from_value: object
        """
        if self.__class__ is CriterionFilterSpecification:
            raise NotImplementedError('Abstract class')
        LeafFilterSpecification.__init__(self)
        self.attr_name = attr_name
        self.attr_value = attr_value

    def __eq__(self, other):
        return (isinstance(other, CriterionFilterSpecification)
                and self.attr_name == other.attr_name
                and self.attr_value == other.attr_value)

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        str_format = '<%s op_name: %s, attr_name: %s, attr_value: %s>'
        params = (self.__class__.__name__,
                  self.operator.name, self.attr_name, self.attr_value)
        return str_format % params

    def is_satisfied_by(self, candidate):
        cand_value = self._get_candidate_value(candidate)
        if IMemberResource.providedBy(self.attr_value): # pylint: disable=E1101
            attr_value = self.attr_value.get_entity()
        elif ICollectionResource.providedBy(self.attr_value): # pylint: disable=E1101
            attr_value = self.attr_value.get_aggregate()
        else:
            attr_value = self.attr_value
        return self.operator.apply(cand_value, attr_value)

    def _get_candidate_value(self, candidate):
        attr_func = get_nested_attribute if '.' in self.attr_name else getattr
        return attr_func(candidate, self.attr_name)


class CompositeFilterSpecification(FilterSpecification):
    """
    Abstract base class for specifications that are composed of two other
    specifications.
    """
    def __init__(self, left_spec, right_spec):
        """
        Constructs a CompositeFilterSpecification

        :param left_spec: the left part of the composite specification
        :type left_spec: :class:`FilterSpecification`
        :param right_spec: the right part of the composite specification
        :type right_spec: :class:`FilterSpecification`
        """
        if self.__class__ is CompositeFilterSpecification:
            raise NotImplementedError('Abstract class')
        FilterSpecification.__init__(self)
        self.__left_spec = left_spec
        self.__right_spec = right_spec

    def __str__(self):
        str_format = '<%s left_spec: %s, right_spec: %s>'
        params = (self.__class__.__name__, self.left_spec, self.right_spec)
        return str_format % params

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.left_spec == other.left_spec
                and self.right_spec == other.right_spec)

    def __ne__(self, other):
        return not self.__eq__(other)

    def accept(self, visitor):
        self.left_spec.accept(visitor)
        self.right_spec.accept(visitor)
        visitor.visit_binary(self)

    @property
    def left_spec(self):
        return self.__left_spec

    @property
    def right_spec(self):
        return self.__right_spec

    def is_satisfied_by(self, candidate):
        return self.operator.apply(self.left_spec.is_satisfied_by(candidate),
                                   self.right_spec.is_satisfied_by(candidate))


class ConjunctionFilterSpecification(CompositeFilterSpecification):
    """
    Concrete conjunction filter specification.
    """
    operator = CONJUNCTION


class DisjunctionFilterSpecification(CompositeFilterSpecification):
    """
    Concrete disjuction filter specification.
    """
    operator = DISJUNCTION


class NegationFilterSpecification(FilterSpecification):
    """
    Concrete negation specification.
    """
    operator = NEGATION

    def __init__(self, wrapped_spec):
        """
        Constructs a NegationFilterSpecification.

        :param wrapped: the wrapped specification
        :type wrapped: :class:`FilterSpecification`
        """
        FilterSpecification.__init__(self)
        self.__wrapped_spec = wrapped_spec

    def __eq__(self, other):
        """Equality operator"""
        return (isinstance(other, NegationFilterSpecification) and
                self.wrapped_spec == other.wrapped_spec)

    def __ne__(self, other):
        """Inequality operator"""
        return not (self == other)

    def __str__(self):
        str_format = '<%s wrapped_spec: %s>'
        params = (self.__class__.__name__, self.wrapped_spec)
        return str_format % params

    def is_satisfied_by(self, candidate):
        return self.operator.apply(
                                self.wrapped_spec.is_satisfied_by(candidate))

    def accept(self, visitor):
        self.wrapped_spec.accept(visitor)
        visitor.visit_unary(self)

    @property
    def wrapped_spec(self):
        """
        Returns the wrapped (negated) specification.
        """
        return self.__wrapped_spec


class ValueStartsWithFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value starts with specification.
    """
    operator = STARTS_WITH


class ValueEndsWithFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value ends with specification.
    """
    operator = ENDS_WITH


class ValueContainsFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value contains specification.
    """
    operator = CONTAINS


class ValueContainedFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value contained in a list of values specification.
    """
    operator = CONTAINED


class ValueEqualToFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value equal to specification.
    """
    operator = EQUAL_TO


class ValueGreaterThanFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value greater than specification.
    """
    operator = GREATER_THAN


class ValueLessThanFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value less than specification.
    """
    operator = LESS_THAN


class ValueGreaterThanOrEqualToFilterSpecification(
                                            CriterionFilterSpecification):
    """
    Concrete value greater than or equal to specification.
    """
    operator = GREATER_OR_EQUALS


class ValueLessThanOrEqualToFilterSpecification(CriterionFilterSpecification):
    """
    Concrete value less than or equal to specification.
    """
    operator = LESS_OR_EQUALS


class ValueInRangeFilterSpecification(CriterionFilterSpecification):
    """
    Concrete specification for a range of values.
    """
    operator = IN_RANGE

    @property
    def from_value(self):
        """
        Returns the first (FROM) value from the range specification.
        """
        return self.attr_value[0]

    @property
    def to_value(self):
        """
        Returns the second (TO) value from the range specification.
        """
        return self.attr_value[1]


@implementer(IFilterSpecificationFactory)
class FilterSpecificationFactory(object):
    """
    Filter specification factory.
    """
    def create_equal_to(self, attr_name, attr_value):
        return ValueEqualToFilterSpecification(attr_name, attr_value)

    def create_starts_with(self, attr_name, attr_value):
        return ValueStartsWithFilterSpecification(attr_name, attr_value)

    def create_ends_with(self, attr_name, attr_value):
        return ValueEndsWithFilterSpecification(attr_name, attr_value)

    def create_contains(self, attr_name, attr_value):
        return ValueContainsFilterSpecification(attr_name, attr_value)

    def create_contained(self, attr_name, attr_value):
        return ValueContainedFilterSpecification(attr_name, attr_value)

    def create_greater_than_or_equal_to(self, attr_name, attr_value):
        return ValueGreaterThanOrEqualToFilterSpecification(attr_name,
                                                            attr_value)

    def create_greater_than(self, attr_name, attr_value):
        return ValueGreaterThanFilterSpecification(attr_name, attr_value)

    def create_less_than_or_equal_to(self, attr_name, attr_value):
        return ValueLessThanOrEqualToFilterSpecification(attr_name, attr_value)

    def create_less_than(self, attr_name, attr_value):
        return ValueLessThanFilterSpecification(attr_name, attr_value)

    def create_in_range(self, attr_name, range_tuple):
        return ValueInRangeFilterSpecification(attr_name, range_tuple)

    def create_conjunction(self, left_spec, right_spec):
        return ConjunctionFilterSpecification(left_spec, right_spec)

    def create_disjunction(self, left_spec, right_spec):
        return DisjunctionFilterSpecification(left_spec, right_spec)

    def create_negation(self, wrapped):
        return NegationFilterSpecification(wrapped)


class OrderSpecification(Specification):
    """
    Abstract base class for all order specifications.
    """
    def __init__(self):
        if self.__class__ is OrderSpecification:
            raise NotImplementedError('Abstract class')
        Specification.__init__(self)

    def eq(self, x, y):
        raise NotImplementedError('Abstract method')

    def lt(self, x, y):
        raise NotImplementedError('Abstract method')

    def le(self, x, y):
        raise NotImplementedError('Abstract method')

    def cmp(self, x, y):
        raise NotImplementedError('Abstract method')

    def ne(self, x, y):
        return not self.eq(x, y)

    def gt(self, x, y):
        return not self.le(x, y)

    def ge(self, x, y):
        return not self.lt(x, y)

    def __and__(self, other):
        return ConjunctionOrderSpecification(self, other)


class ObjectOrderSpecification(OrderSpecification): # pylint: disable=W0223
    """
    Abstract base class for all order specifications operating on object
    attributes.
    """
    def __init__(self, attr_name):
        if self.__class__ is ObjectOrderSpecification:
            raise NotImplementedError('Abstract class')
        OrderSpecification.__init__(self)
        self.__attr_name = attr_name
        self.__attr_func = \
                get_nested_attribute if '.' in attr_name else getattr

    def __str__(self):
        str_format = '<%s attr_name: %s>'
        params = (self.__class__.__name__, self.attr_name)
        return str_format % params

    @property
    def attr_name(self):
        return self.__attr_name

    def eq(self, x, y):
        res = self.operator.apply(self._get_value(x), self._get_value(y))
        return res == 0

    def lt(self, x, y):
        res = self.operator.apply(self._get_value(x), self._get_value(y))
        return res == -1

    def le(self, x, y):
        res = self.operator.apply(self._get_value(x), self._get_value(y))
        return res == -1 or res == 0

    def cmp(self, x, y):
        return self.operator.apply(self._get_value(x), self._get_value(y))

    def accept(self, visitor):
        visitor.visit_nullary(self)

    def _get_value(self, obj):
        return self.__attr_func(obj, self.attr_name)


class AscendingOrderSpecification(ObjectOrderSpecification):
    """
    Concrete ascending order specification.
    """
    operator = ASCENDING


class DescendingOrderSpecification(ObjectOrderSpecification):
    """
    Concrete descending order specification.
    """
    operator = DESCENDING


class NaturalOrderSpecification(ObjectOrderSpecification):
    """
    See http://www.codinghorror.com/blog/2007/12/sorting-for-humans-natural-sort-order.html
    """
    operator = ASCENDING

    def _get_value(self, obj):
        value = ObjectOrderSpecification._get_value(self, obj)
        if isinstance(value, string_types):
            res = [self.__convert(c) for c in re.split(r'([0-9]+)', value)]
        else:
            res = value
        return res

    def __convert(self, txt):
        return int(txt) if txt.isdigit() else txt


class ConjunctionOrderSpecification(OrderSpecification):
    """
    Concrete conjunction order specification.
    """
    operator = CONJUNCTION

    def __init__(self, left, right):
        OrderSpecification.__init__(self)
        self.__left = left
        self.__right = right

    def __str__(self):
        str_format = '<%s left_spec: %s, right_spec: %s>'
        params = (self.__class__.__name__, self.left, self.right)
        return str_format % params

    def eq(self, x, y):
        return self.left.eq(x, y) and self.right.eq(x, y)

    def lt(self, x, y):
        if self.left.eq(x, y):
            res = self.right.lt(x, y)
        else:
            res = self.left.lt(x, y)
        return res

    def le(self, x, y):
        if self.left.eq(x, y):
            res = self.right.le(x, y)
        else:
            res = self.left.le(x, y)
        return res

    def cmp(self, x, y):
        left_cmp = self.left.cmp(x, y)
        if left_cmp == 0:
            res = self.right.cmp(x, y)
        else:
            res = left_cmp
        return res

    @property
    def left(self):
        return self.__left

    @property
    def right(self):
        return self.__right

    def accept(self, visitor):
        self.__left.accept(visitor)
        self.__right.accept(visitor)
        visitor.visit_binary(self)


@implementer(IOrderSpecificationFactory)
class OrderSpecificationFactory(object):
    """
    Order specification factory.
    """
    def create_ascending(self, attr_name):
        return AscendingOrderSpecification(attr_name)

    def create_descending(self, attr_name):
        return DescendingOrderSpecification(attr_name)

    def create_natural(self, attr_name):
        return NaturalOrderSpecification(attr_name)

    def create_conjunction(self, left_spec, right_spec):
        return ConjunctionOrderSpecification(left_spec, right_spec)


class specification_attribute(object):
    """
    Helper descriptor for the :class:`SpecificationGenerator`.
    """
    def __init__(self, ifactory, method_name):
        self.__method_name = method_name
        self.__ifactory = ifactory
        self.__factory = None

    def __get__(self, generator, generator_class):
        if generator is None:
            generator = generator_class(self.factory)
        generator.method_name = self.__method_name
        return generator

    @property
    def factory(self):
        if self.__factory is None:
            reg = get_current_registry()
            self.__factory = reg.queryUtility(self.__ifactory)
        return self.__factory


class _SpecificationGenerator(object):
    """
    Base class for specification generators.
    """
    def __init__(self, factory):
        self._factory = factory
        self.method_name = None


class FilterSpecificationGenerator(_SpecificationGenerator):
    """
    Helper class to simplify the generation of filter specifications.
    """
    eq = specification_attribute(IFilterSpecificationFactory,
                                 'create_equal_to')
    starts = specification_attribute(IFilterSpecificationFactory,
                                     'create_starts_with')
    ends = specification_attribute(IFilterSpecificationFactory,
                                   'create_ends_with')
    lt = specification_attribute(IFilterSpecificationFactory,
                                 'create_less_than')
    le = specification_attribute(IFilterSpecificationFactory,
                                 'create_less_than_or_equal_to')
    gt = specification_attribute(IFilterSpecificationFactory,
                                 'create_greater_than')
    ge = specification_attribute(IFilterSpecificationFactory,
                                 'create_greater_than_or_equal_to')
    cnts = specification_attribute(IFilterSpecificationFactory,
                                   'create_contains')
    cntd = specification_attribute(IFilterSpecificationFactory,
                                   'create_contained')
    rng = specification_attribute(IFilterSpecificationFactory,
                                  'create_in_range')

    def __call__(self, **kw):
        fn = getattr(self._factory, self.method_name)
        spec = None
        for (attr, value) in kw.items():
            if spec is None:
                spec = fn(attr, value)
            else:
                spec = spec & fn(attr, value)
        return spec


class SingleOrderSpecificationGenerator(_SpecificationGenerator):
    """
    Helper class to simplify the generation of order specifications.
    """
    asc = specification_attribute(IOrderSpecificationFactory,
                                  'create_ascending')
    desc = specification_attribute(IOrderSpecificationFactory,
                                   'create_descending')

    def __call__(self, *args):
        fn = getattr(self._factory, self.method_name)
        spec = None
        for attr in args:
            if spec is None:
                spec = fn(attr)
            else:
                spec = spec & (fn(attr))
        return spec


class GenericOrderSpecificationGenerator(_SpecificationGenerator):
    """
    Helper class to simplify the generation of generic order specifications.
    """
    order = specification_attribute(IOrderSpecificationFactory, None)

    def __call__(self, *args):
        spec = None
        for order_crit in args:
            name, order_op = order_crit
            if order_op == ASCENDING:
                fn = self._factory.create_ascending
            elif order_op == DESCENDING:
                fn = self._factory.create_descending
            else:
                raise ValueError('Invalid ordering operator "%s".' % order_op)
            item_spec = fn(name)
            if spec is None:
                spec = item_spec
            else:
                spec &= item_spec
        return spec


def eq(**kw):
    "Convenience function to create an equal_to specification."
    return FilterSpecificationGenerator.eq(**kw)


def starts(**kw):
    "Convenience function to create a starts_with specification."
    return FilterSpecificationGenerator.starts(**kw)


def ends(**kw):
    "Convenience function to create an ends_with specification."
    return FilterSpecificationGenerator.ends(**kw)


def lt(**kw):
    "Convenience function to create a less_than specification."
    return FilterSpecificationGenerator.lt(**kw)


def le(**kw):
    "Convenience function to create less_than_or_equal_to specification."
    return FilterSpecificationGenerator.le(**kw)


def gt(**kw):
    "Convenience function to create a greater_than specification."
    return FilterSpecificationGenerator.gt(**kw)


def ge(**kw):
    "Convenience function to create a greater_than_or_equal specification."
    return FilterSpecificationGenerator.ge(**kw)


def cnts(**kw):
    "Convenience function to create a contains specification."
    return FilterSpecificationGenerator.cnts(**kw)


def cntd(**kw):
    "Convenience function to create a contained specification."
    return FilterSpecificationGenerator.cntd(**kw)


def rng(**kw):
    "Convenience function to create an in_range specification."
    return FilterSpecificationGenerator.rng(**kw)


def asc(*args):
    "Convenience function to create an ascending order specification."
    return SingleOrderSpecificationGenerator.asc(*args)


def desc(*args):
    "Convenience function to create a descending order specification."
    return SingleOrderSpecificationGenerator.desc(*args)


def order(*args):
    "Convenience function to create an order specification."
    return GenericOrderSpecificationGenerator.order(*args)



"""
URL <-> resource conversion.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 28, 2011.
"""
from everest.compat import parse_qsl
from everest.interfaces import IResourceUrlConverter
from everest.querying.base import EXPRESSION_KINDS
from everest.querying.filterparser import parse_filter
from everest.querying.orderparser import parse_order
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IResource
from everest.resources.utils import get_root_collection
from everest.utils import get_filter_specification_visitor
from everest.utils import get_order_specification_visitor
from pyparsing import ParseException
from pyramid.compat import url_unquote
from pyramid.compat import urlparse
from pyramid.traversal import find_resource
from pyramid.traversal import traversal_path
from zope.interface import implementer # pylint: disable=E0611,F0401
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from everest.querying.refsparser import parse_refs

__docformat__ = 'reStructuredText en'
__all__ = ['ResourceUrlConverter',
           'UrlPartsConverter',
           ]


@implementer(IResourceUrlConverter)
class ResourceUrlConverter(object):
    """
    Performs URL <-> resource instance conversion.

    See http://en.wikipedia.org/wiki/Query_string for information on characters
    supported in query strings.
    """

    def __init__(self, request):
        # The request is needed for access to app URL, registry, traversal.
        self.__request = request

    def url_to_resource(self, url):
        """
        Returns the resource that is addressed by the given URL.

        :param str url: URL to convert
        :return: member or collection resource

        :note: If the query string in the URL has multiple values for a
          query parameter, the last definition in the query string wins.
        """
        parsed = urlparse.urlparse(url)
        parsed_path = parsed.path # namedtupble problem pylint: disable=E1101
        rc = find_resource(self.__request.root, traversal_path(parsed_path))
        if ICollectionResource in provided_by(rc):
            # In case we found a collection, we have to filter, order, slice.
            parsed_query = parsed.query # namedtuple problem pylint: disable=E1101
            params = dict(parse_qsl(parsed_query))
            filter_string = params.get('q')
            if not filter_string is None:
                rc.filter = \
                    UrlPartsConverter.make_filter_specification(filter_string)
            order_string = params.get('sort')
            if not order_string is None:
                rc.order = \
                    UrlPartsConverter.make_order_specification(order_string)
            start_string = params.get('start')
            size_string = params.get('size')
            if not (start_string is None or size_string is None):
                rc.slice = \
                  UrlPartsConverter.make_slice_key(start_string, size_string)
        elif not IMemberResource in provided_by(rc):
            raise ValueError('Traversal found non-resource object "%s".' % rc)
        return rc

    def resource_to_url(self, resource, quote=False):
        """
        Returns the URL for the given resource.

        :param resource: Resource to create a URL for.
        :param bool quote: If set, the URL returned will be quoted.
        :raises ValueError: If the given resource is floating (i.e., has
          the parent attribute set to `None`)
        """
        ifc = provided_by(resource)
        if not IResource in ifc:
            raise TypeError('Can not generate URL for non-resource "%s".'
                            % resource)
        elif resource.__parent__ is None:
            raise ValueError('Can not generate URL for floating resource '
                             '"%s".' % resource)
        if ICollectionResource in ifc:
            query = {}
            if not resource.filter is None:
                query['q'] = \
                    UrlPartsConverter.make_filter_string(resource.filter)
            if not resource.order is None:
                query['sort'] = \
                    UrlPartsConverter.make_order_string(resource.order)
            if not resource.slice is None:
                query['start'], query['size'] = \
                    UrlPartsConverter.make_slice_strings(resource.slice)
            if query != {}:
                options = dict(query=query)
            else:
                options = dict()
            if not resource.is_root_collection:
                # For nested collections, we check if the referenced root
                # collection is exposed (i.e., has the service as parent).
                # If yes, we return an absolute URL, else a nested URL.
                root_coll = get_root_collection(resource)
                if not root_coll.has_parent:
                    url = self.__request.resource_url(resource)
                else:
                    url = self.__request.resource_url(root_coll, **options)
            else:
                url = self.__request.resource_url(resource, **options)
        else:
            if not resource.is_root_member:
                # For nested members, we check if the referenced root
                # collection is exposed (i.e., has the service as parent).
                # If yes, we return an absolute URL, else a nested URL.
                root_coll = get_root_collection(resource)
                if not root_coll.has_parent:
                    url = self.__request.resource_url(resource)
                else:
                    par_url = self.__request.resource_url(root_coll)
                    url = "%s%s/" % (par_url, resource.__name__)
            else:
                url = self.__request.resource_url(resource)
        if not quote:
            url = url_unquote(url)
        return url


class UrlPartsConverter(object):
    """
    Helper class providing functionality to convert parts of a URL to
    specifications and vice versa.
    """
    @classmethod
    def make_filter_specification(cls, filter_string):
        """
        Converts the given CQL filter expression into a filter specification.
        """
        try:
            return parse_filter(filter_string)
        except ParseException as err:
            raise ValueError('Expression parameters have errors. %s' % err)

    @classmethod
    def make_filter_string(cls, filter_specification):
        """
        Converts the given filter specification to a CQL filter expression.
        """
        visitor_cls = get_filter_specification_visitor(EXPRESSION_KINDS.CQL)
        visitor = visitor_cls()
        filter_specification.accept(visitor)
        return str(visitor.expression)

    @classmethod
    def make_order_specification(cls, order_string):
        """
        Converts the given CQL sort expression to a order specification.
        """
        try:
            return parse_order(order_string)
        except ParseException as err:
            raise ValueError('Expression parameters have errors. %s' % err)

    @classmethod
    def make_order_string(cls, order_specification):
        """
        Converts the given order specification to a CQL order expression.
        """
        visitor_cls = get_order_specification_visitor(EXPRESSION_KINDS.CQL)
        visitor = visitor_cls()
        order_specification.accept(visitor)
        return str(visitor.expression)

    @classmethod
    def make_slice_key(cls, start_string, size_string):
        """
        Converts the given start and size query parts to a slice key.

        :return: slice key
        :rtype: slice
        """
        try:
            start = int(start_string)
        except ValueError:
            raise ValueError('Query parameter "start" must be a number.')
        if start < 0:
            raise ValueError('Query parameter "start" must be zero or '
                             'a positive number.')
        try:
            size = int(size_string)
        except ValueError:
            raise ValueError('Query parameter "size" must be a number.')
        if size < 1:
            raise ValueError('Query parameter "size" must be a positive '
                             'number.')
        return slice(start, start + size)

    @classmethod
    def make_slice_strings(cls, slice_key):
        """
        Converts the given slice key to start and size query parts.
        """
        start = slice_key.start
        size = slice_key.stop - start
        return (str(start), str(size))

    @classmethod
    def make_refs_options(cls, refs_string):
        """
        Converts the given CQL resource references string to a dictionary of
        attribute representer options.
        """
        try:
            return parse_refs(refs_string)
        except ParseException as err:
            raise ValueError('Refs string has errors. %s' % err)



"""
Entity tree traversal operations.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 8, 2013.
"""
from everest.constants import RELATIONSHIP_DIRECTIONS
from everest.constants import RELATION_OPERATIONS
from everest.entities.attributes import get_domain_class_attribute_iterator
from everest.entities.relationship import LazyDomainRelationship
from everest.entities.utils import get_entity_class
from everest.interfaces import IDataTraversalProxyFactory
from everest.resources.interfaces import ICollectionResource
from everest.resources.utils import url_to_resource
from everest.traversal import DataTraversalProxy
from everest.traversal import DataTraversalProxyAdapter
from everest.utils import get_nested_attribute
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['AruVisitor',
           'DomainDataTraversalProxy',
           'DomainDataTraversalProxyAdapter',
           'LinkedDomainDataTraversalProxy',
           'LinkedDomainDataTraversalProxyAdapter',
           'ResourceDataTreeVisitor',
           ]


class ResourceDataTreeVisitor(object):
    """
    Abstract base class for resource data tree traversal visitors.
    """
    def __init__(self, rc_class):
        self._rc_class = rc_class

    def prepare(self):
        """
        Prepare the visitor for traversal.

        Called by the traverser; override to implement functionality needed
        to prepare the visitor for a new resource tree traversal operation.
        """
        pass #pragma: no cover

    def finalize(self):
        """
        Finalize the visitor after traversal.

        Called by the traverser; override to implement functionality the
        visitor needs to run ''after'' traversal of the full resource tree
        has finished.
        """
        pass #pragma: no cover

    def visit(self, path, attribute, source, target):
        """
        Visits a single node in the resource data tree.

        The relation operation for the visit can be determined implicitly from
        the values of the :param:`source` and :param:`target` parameters:
        REMOVE if only :param:`source` is `None`, ADD if only :param:`target`,
        and UPDATE if both are not `None`.

        :param path: Traversal path.
        :param attribute: Resource attribute to traverse. This is `None` for
          nodes at the root of the tree.
        :param source: Traversal proxy for source data.
        :param target: Traversal proxy for target data.
        """
        raise NotImplementedError('Abstract method.')


class AruVisitor(ResourceDataTreeVisitor):
    """
    Add-Remove-Update visitor for resource data tree traversal.
    """
    class Callback(object):
        operation = None
        def __init__(self, method, args):
            self.method = method
            self.args = args

        def __call__(self):
            self.method(*self.args)

        def __str__(self):
            return "%s: %s (%d)" \
                   % (self.operation, self.args[1], len(self.args[-1])) #pragma: no cover

    class AddCallback(Callback):
        operation = 'ADD'

    class RemoveCallback(Callback):
        operation = 'REMOVE'

    class UpdateCallback(Callback):
        operation = 'UPDATE'

    def __init__(self, rc_class, add_callback=None, remove_callback=None,
                 update_callback=None, pass_path_to_callbacks=False):
        ResourceDataTreeVisitor.__init__(self, rc_class)
        self.__add_callback = add_callback
        self.__remove_callback = remove_callback
        self.__update_callback = update_callback
        self.__pass_path_to_callbacks = pass_path_to_callbacks
        self.__commands = None
        self.root = None

    def prepare(self):
        #: The root of the new source tree (ADD) or of the updated target
        #: tree (UPDATE) or the removed entity (REMOVE).
        self.root = None
        #
        self.__commands = []

    def visit(self, path, attribute, source, target):
        is_root = attribute is None
        if is_root:
            # Visiting the root.
            ent_class = get_entity_class(self._rc_class)
        else:
            ent_class = get_entity_class(attribute.attr_type)
            parent = path.parent
        if source is None:
            # No source - REMOVE.
            entity = target.get_entity()
            if not is_root:
                rel = self.__get_relationship(parent, attribute)
                rel.remove(entity)
            if not self.__remove_callback is None:
                if self.__pass_path_to_callbacks:
                    args = (ent_class, entity, path)
                else:
                    args = (ent_class, entity)
                cmd = self.RemoveCallback(self.__remove_callback, args)
                self.__commands.append(cmd)
        else:
            if target is None:
                # No target - ADD.
                entity = source.get_entity()
                if not is_root:
                    if path.relation_operation == RELATION_OPERATIONS.ADD:
                        # If the parent is created new, the constructor
                        # will most likely set the child attribute.
                        add_opts = dict(safe=True)
                    else:
                        add_opts = dict()
                    rel = self.__get_relationship(parent, attribute)
                    rel.add(entity, **add_opts)
                if not self.__add_callback is None:
                    if self.__pass_path_to_callbacks:
                        args = (ent_class, entity, path)
                    else:
                        args = (ent_class, entity)
                    cmd = self.AddCallback(self.__add_callback, args)
                    self.__commands.append(cmd)
            else:
                # Both source and target - UPDATE.
                entity = target.get_entity()
                if not self.__update_callback is None:
                    upd_av_map = dict(source.update_attribute_value_items)
                    if self.__pass_path_to_callbacks:
                        args = (ent_class, upd_av_map, entity, path)
                    else:
                        args = (ent_class, upd_av_map, entity)
                    cmd = self.UpdateCallback(self.__update_callback, args)
                    self.__commands.append(cmd)
                if not is_root:
                    # The relationship with the old value has already been
                    # severed, so we only need to ADD here.
                    rel = self.__get_relationship(parent, attribute)
                    rel.add(entity)
        if is_root:
            self.root = entity
        else:
            self.__commands.append(rel)

    def finalize(self):
        for cmd in self.__commands:
            cmd()

    def __get_relationship(self, proxy, attribute):
        try:
            rel = proxy.get_relationship(attribute)
        except KeyError:
            rel = LazyDomainRelationship(
                                proxy, attribute,
                                direction=proxy.relationship_direction)
            proxy.set_relationship(attribute, rel)
        return rel


class DomainDataTraversalProxy(DataTraversalProxy):
    def get_id(self):
        return self._data.id

    def _get_entity_type(self):
        return type(self._data)

    def get_entity(self):
        return self._data

    def _attribute_iterator(self):
        it = get_domain_class_attribute_iterator(self._data)
        for attr in it:
            if not attr.entity_attr is None:
                yield attr

    def _get_relation_attribute_value(self, attribute):
        return get_nested_attribute(self._data, attribute.entity_attr)

    def _get_proxied_attribute_value(self, attribute):
        return get_nested_attribute(self._data, attribute.entity_attr)

    def _make_accessor(self, value_type):
        return self._accessor.get_root_aggregate(value_type)


class DomainDataTraversalProxyAdapter(DataTraversalProxyAdapter):
    proxy_class = DomainDataTraversalProxy

    def make_source_proxy(self, options=None):
        rel_drct = RELATIONSHIP_DIRECTIONS.NONE
        return self.make_proxy(None, rel_drct, options=options)

    def make_target_proxy(self, accessor,
                          manage_back_references=True, options=None):
        rel_drct = RELATIONSHIP_DIRECTIONS.BIDIRECTIONAL
        if not manage_back_references:
            rel_drct &= ~RELATIONSHIP_DIRECTIONS.REVERSE
        return self.make_proxy(accessor, rel_drct, options=options)


class LinkedDomainDataTraversalProxy(DomainDataTraversalProxy):
    def do_traverse(self):
        return False

    def __str__(self):
        return "Linked%s" % DomainDataTraversalProxy.__str__(self)


class LinkedDomainDataTraversalProxyAdapter(DomainDataTraversalProxyAdapter):
    proxy_class = LinkedDomainDataTraversalProxy

    def make_target_proxy(self, accessor,
                          manage_back_references=True, options=None):
        raise NotImplementedError('Not implemented.')

    def make_proxy(self, accessor, relationship_direction, options=None):
        # Note: We ignore the options; if something was passed here,
        #       it came from a parent data element data traversal proxy.
        url = self._data.get_url()
        rc = url_to_resource(url)
        if ICollectionResource.providedBy(rc): # pylint:disable=E1101
            reg = get_current_registry()
            prx_fac = reg.getUtility(IDataTraversalProxyFactory)
            prx = prx_fac.make_proxy([mb.get_entity() for mb in rc],
                                     accessor, relationship_direction)
        else:
            prx = LinkedDomainDataTraversalProxy(rc.get_entity(),
                                                 accessor,
                                                 relationship_direction)
        return prx




"""
Custom querying operators.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 5, 2011.
"""
from pyramid.compat import string_types

__docformat__ = 'reStructuredText en'
__all__ = ['ASCENDING',
           'BinaryOperator',
           'CONJUNCTION',
           'CONTAINED',
           'CONTAINS',
           'CQL_FILTER_OPERATORS',
           'CQL_ORDER_OPERATORS',
           'DESCENDING',
           'DISJUNCTION',
           'ENDS_WITH',
           'EQUAL_TO',
           'GREATER_OR_EQUALS',
           'GREATER_THAN',
           'IN_RANGE',
           'LESS_OR_EQUALS',
           'LESS_THAN',
           'NEGATION',
           'NullaryOperator',
           'Operator',
           'STARTS_WITH',
           'UnaryOperator',
           ]


class Operator(object):
    """
    Base class for querying operators.
    """
    #: The name of the operator. To be specified in derived classes.
    name = None
    #: The arity (number of arguments required) of the operator. To be
    #: specified in derived classes.
    arity = None


class NullaryOperator(Operator):
    """
    Nullary querying operator.
    """
    arity = 0

    @staticmethod
    def apply():
        raise NotImplementedError('Abstract method.')


class UnaryOperator(Operator):
    """
    Unary querying operator.
    """
    arity = 1

    @staticmethod
    def apply(value):
        raise NotImplementedError('Abstract method.')


class BinaryOperator(Operator):
    """
    Binary querying operator.
    """
    arity = 2

    @staticmethod
    def apply(value, ref_value):
        raise NotImplementedError('Abstract method.')


class STARTS_WITH(BinaryOperator):
    name = 'starts_with'

    @staticmethod
    def apply(value, ref_value):
        if isinstance(value, string_types):
            res = value.startswith(ref_value)
        else:
            res = value[0] == ref_value
        return res


class ENDS_WITH(BinaryOperator):
    name = 'ends_with'

    @staticmethod
    def apply(value, ref_value):
        if isinstance(value, string_types):
            res = value.endswith(ref_value)
        else:
            res = value[-1] == ref_value
        return res


class CONTAINED(BinaryOperator):
    name = 'contained'

    @staticmethod
    def apply(value, ref_value):
        return value in ref_value


class CONTAINS(BinaryOperator):
    name = 'contains'

    @staticmethod
    def apply(value, ref_value):
        return ref_value in value


class EQUAL_TO(BinaryOperator):
    name = 'equal_to'

    @staticmethod
    def apply(value, ref_value):
        return value == ref_value


class LESS_THAN(BinaryOperator):
    name = 'less_than'

    @staticmethod
    def apply(value, ref_value):
        return value < ref_value


class LESS_OR_EQUALS(BinaryOperator):
    name = 'less_than_or_equal_to'

    @staticmethod
    def apply(value, ref_value):
        return value <= ref_value


class GREATER_THAN(BinaryOperator):
    name = 'greater_than'

    @staticmethod
    def apply(value, ref_value):
        return value > ref_value


class GREATER_OR_EQUALS(BinaryOperator):
    name = 'greater_than_or_equal_to'

    @staticmethod
    def apply(value, ref_value):
        return value >= ref_value


class IN_RANGE(BinaryOperator):
    name = 'in_range'

    @staticmethod
    def apply(value, ref_value):
        return value >= ref_value[0] and value <= ref_value[1]


class ASCENDING(BinaryOperator):
    name = 'asc'

    @staticmethod
    def apply(x, y):
        return (x > y) - (x < y) # PY3 compatible cmp replacement.


class DESCENDING(BinaryOperator):
    name = 'desc'

    @staticmethod
    def apply(x, y):
        return (y > x) - (y < x) # PY3 compatible cmp replacement.


class NEGATION(UnaryOperator):
    name = 'negation'

    @staticmethod
    def apply(x):
        return not x


class CONJUNCTION(BinaryOperator):
    name = 'conjunction'

    @staticmethod
    def apply(x, y):
        return x and y


class DISJUNCTION(BinaryOperator):
    name = 'disjunction'

    @staticmethod
    def apply(x, y):
        return x or y


class CQL_FILTER_OPERATORS(object):
    """
    Static container for all CQL filtering operators.
    """
    STARTS_WITH = STARTS_WITH
    ENDS_WITH = ENDS_WITH
    CONTAINED = CONTAINED
    CONTAINS = CONTAINS
    EQUAL_TO = EQUAL_TO
    LESS_THAN = LESS_THAN
    LESS_OR_EQUALS = LESS_OR_EQUALS
    GREATER_THAN = GREATER_THAN
    GREATER_OR_EQUALS = GREATER_OR_EQUALS
    IN_RANGE = IN_RANGE


class CQL_ORDER_OPERATORS(object):
    """
    Static container for all CQL ordering operators.
    """
    ASCENDING = ASCENDING
    DESCENDING = DESCENDING



from everest.resources.base import Collection
from everest.resources.base import Member


class FooMember(Member):
    relation = 'http://everest.org/relations/foomember'


class BarMember(Member):
    relation = 'http://everest.org/relations/barmember'


class FooCollection(Collection):
    title = 'My Foo Collection'
    root_name = 'foos'
    description = 'My fancy Foo collection.'
    max_limit = 1000


class BarCollection(Collection):
    title = 'My Bar Collection'
    root_name = 'bars'
    description = 'My fancy Bar collection.'



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 12, 2013.
"""
from everest.constants import RELATION_OPERATIONS
from everest.entities.attributes import get_domain_class_attribute
from everest.entities.base import Entity
from everest.entities.traversal import AruVisitor
from everest.entities.traversal import DomainDataTraversalProxy
from everest.repositories.memory.cache import EntityCacheMap
from everest.resources.staging import StagingAggregate
from everest.resources.staging import create_staging_collection
from everest.testing import EntityTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.testing import create_entity
from everest.traversal import SourceTargetDataTreeTraverser
from mock import MagicMock

__docformat__ = 'reStructuredText en'
__all__ = ['SourceTargetDataTraverserTestCase',
           ]


class SourceTargetDataTraverserTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_traverse_with_add(self):
        mock_vst = MagicMock()
        ent = create_entity(entity_id=None)
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                    ent, None,
                                                    RELATION_OPERATIONS.ADD)
        trv.run(mock_vst)
        parent_attr = get_domain_class_attribute(MyEntity, 'parent')
        children_attr = get_domain_class_attribute(MyEntity, 'children')
        grandchildren_attr = get_domain_class_attribute(MyEntityChild,
                                                        'children')
        # Check the visiting sequence and args (depth first).
        for idx, (meth_name, attr) \
            in enumerate([('visit', parent_attr),
                          ('visit', grandchildren_attr),
                          ('visit', children_attr),
                          ]):
            meth_call = mock_vst.method_calls[idx + 1]
            self.assert_equal(meth_call[0], meth_name)
            self.assert_equal(meth_call[1][1], attr)
            prx = meth_call[1][2]
            self.assert_true(isinstance(prx, DomainDataTraversalProxy))
            self.assert_true(isinstance(prx.get_entity(), Entity))
            self.assert_is_none(meth_call[1][3])

    def test_make_traverser_invalid_params(self):
        ent0 = create_entity(entity_id=None)
        ent1 = create_entity(entity_id=None)
        self.assert_raises(ValueError,
                           SourceTargetDataTreeTraverser.make_traverser,
                           ent0, ent1,
                           RELATION_OPERATIONS.ADD)
        self.assert_raises(ValueError,
                           SourceTargetDataTreeTraverser.make_traverser,
                           ent1, ent0,
                           RELATION_OPERATIONS.REMOVE)
        self.assert_raises(ValueError,
                           SourceTargetDataTreeTraverser.make_traverser,
                           ent0, None,
                           RELATION_OPERATIONS.UPDATE,
                           accessor=None)

    def test_make_traverser_update(self):
        ent0 = create_entity(entity_id=0)
        ent1 = create_entity(entity_id=None)
        agg = create_staging_collection(IMyEntity).get_aggregate()
        agg.add(ent0)
        agg.add(ent1)
        ent01 = create_entity(entity_id=0)
        ent11 = create_entity(entity_id=None)
        # With many as source and one as target.
        with self.assert_raises(ValueError) as cm:
            SourceTargetDataTreeTraverser.make_traverser(
                                                [ent01, ent1], ent0,
                                                RELATION_OPERATIONS.UPDATE,
                                                accessor=agg)
        self.assert_true(
                cm.exception.args[0].endswith('or both not be sequences.'))
        # Without target.
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                [ent01, ent11], None,
                                                RELATION_OPERATIONS.UPDATE,
                                                accessor=agg)
        self.assert_is_not_none(getattr(trv, '_tgt_prx'))

    def test_traverse_with_remove_sequence(self):
        ent0 = create_entity(entity_id=0)
        ent1 = create_entity(entity_id=None)
        cache = EntityCacheMap()
        agg = StagingAggregate(MyEntity, cache=cache)
        agg.add(ent0)
        agg.add(ent1)
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                                None, [ent0, ent1],
                                                RELATION_OPERATIONS.REMOVE)
        vst = AruVisitor(MyEntity, remove_callback=cache.remove)
        trv.run(vst)
        self.assert_equal(len(list(iter(agg))), 0)



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 19, 2012.
"""
from everest.entities.system import UserMessage
from everest.interfaces import IUserMessageNotifier
from everest.messaging import UserMessageChecker
from everest.messaging import UserMessageHandlingContextManager
from everest.messaging import UserMessageNotifier
from everest.testing import TestCaseWithConfiguration

__docformat__ = 'reStructuredText en'
__all__ = ['MessagingTestCase',
           ]


class MessagingTestCase(TestCaseWithConfiguration):
    def set_up(self):
        TestCaseWithConfiguration.set_up(self)
        reg = self.config.registry
        reg.registerUtility(UserMessageNotifier(), # pylint:disable=E1103
                            IUserMessageNotifier)
        self.values = []

    def test_user_message(self):
        self.assert_raises(ValueError, UserMessage, 'blah', guid='-1', id=1)

    def test_single(self):
        checker1 = MyChecker1()
        self.assert_true(checker1.vote is True)
        # Sending with TRUE message triggers an append to self.values.
        with UserMessageHandlingContextManager(checker1):
            self.__routine_which_sends_user_message(MESSAGE_ONE)
        self.assert_equal(self.values, [True])
        self.assert_equal(checker1.vote, True)
        # Sending with "FALSE" message vetoes the append.
        self.values.pop()
        with UserMessageHandlingContextManager(checker1):
            self.__routine_which_sends_user_message(MESSAGE_TWO)
        self.assert_equal(self.values, [])
        self.assert_equal(checker1.vote, False)
        # Sending without registered checker1 triggers an append (because
        # there is no veto), but does not change the checking vote.
        self.__routine_which_sends_user_message(MESSAGE_ONE)
        self.assert_equal(self.values, [True])
        self.assert_equal(checker1.vote, False)

    def test_nested(self):
        checker1 = MyChecker1()
        checker2 = MyChecker2()
        # No action because checker 2 returns False.
        with UserMessageHandlingContextManager(checker1):
            with UserMessageHandlingContextManager(checker2):
                self.__routine_which_sends_user_message(MESSAGE_ONE)
        self.assert_equal(self.values, [])
        self.assert_equal(checker1.vote, False)
        self.assert_equal(checker2.vote, False)
        self.assert_equal(checker1.myvote, True)
        self.assert_equal(checker2.myvote, False)
        # No action because checker 1 returns False. Checker 2 is not touched.
        checker2.myvote = -1
        with UserMessageHandlingContextManager(checker1):
            with UserMessageHandlingContextManager(checker2):
                self.__routine_which_sends_user_message(MESSAGE_TWO)
        self.assert_equal(self.values, [])
        self.assert_equal(checker1.vote, False)
        self.assert_equal(checker2.vote, False)
        self.assert_equal(checker1.myvote, False)
        self.assert_equal(checker2.myvote, -1)
        # Action because checker 1 returns None. Checker 2 is not touched.
        with UserMessageHandlingContextManager(checker1):
            with UserMessageHandlingContextManager(checker2):
                self.__routine_which_sends_user_message(MESSAGE_THREE)
        self.assert_equal(self.values, [True])
        self.assert_equal(checker1.vote, True)
        self.assert_equal(checker2.vote, True)
        self.assert_is_none(checker1.myvote)
        self.assert_equal(checker2.myvote, -1)

    def __routine_which_sends_user_message(self, msg):
        msg_notifier = \
          self.config.get_registered_utility(IUserMessageNotifier)
        if msg_notifier.notify(msg):
            self.values.append(True)


MESSAGE_ONE = 'ONE'
MESSAGE_TWO = 'TWO'
MESSAGE_THREE = 'THREE'


class MyChecker1(UserMessageChecker):
    myvote = -1
    def check(self):
        if self.message.text == MESSAGE_ONE:
            vote = True
        elif self.message.text == MESSAGE_TWO:
            vote = False
        else:
            vote = None
        self.myvote = vote
        return vote


class MyChecker2(UserMessageChecker):
    myvote = -1
    def check(self):
        if self.message.text == MESSAGE_ONE:
            vote = False
        elif self.message.text == MESSAGE_TWO:
            vote = True
        else:
            vote = False
        self.myvote = vote
        return vote



"""
Settings file handling and nose plugin for testing.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 30, 2012.
"""
from pyramid.compat import NativeIO
from pyramid.compat import configparser
import nose.plugins
import os

__docformat__ = 'reStructuredText en'
__all__ = ['EverestIni',
           'EverestNosePlugin',
           ]


class EverestNosePlugin(nose.plugins.Plugin):
    """
    Nose plugin extension.

    Provides a nose option that configures a test application configuration
    file.
    """
    def __init__(self, *args, **kw):
        nose.plugins.Plugin.__init__(self, *args, **kw)
        self.__opt_name = 'app-ini-file'
        self.__dest_opt_name = self.__opt_name.replace('-', '_')

    def options(self, parser, env=None):
        """
        Adds command-line options for this plugin.
        """
        if env is None:
            env = os.environ
        env_opt_name = 'NOSE_%s' % self.__dest_opt_name.upper()
        parser.add_option("--%s" % self.__opt_name,
                          dest=self.__dest_opt_name,
                          type="string",
                          default=env.get(env_opt_name),
                          help=".ini file providing the environment for the "
                               "test web application.")

    def configure(self, options, conf):
        """
        Configures the plugin.
        """
        super(EverestNosePlugin, self).configure(options, conf)
        opt_val = getattr(options, self.__dest_opt_name, None)
        if opt_val:
            self.enabled = True
            EverestIni.ini_file_path = opt_val


class EverestIni(object):
    """
    Helper class providing access to settings parsed from an ini file.

    By default, the ini file configured through the :class:`EverestNosePlugin`
    is used.
    """
    #: Path to the global ini file. Set through the nose plugin.
    ini_file_path = None

    __ini_parser = None

    def __init__(self, ini_file_path=None):
        if ini_file_path is None:
            # Revert to the .ini configured through the --app-ini-file option
            # or initialize empty.
            self.__ini_parser = self.__check_ini_file()
        else:
            self.__ini_parser = self.__make_parser()
            self.__ini_parser.read(ini_file_path)
            self.ini_file_path = ini_file_path

    def get_settings(self, section):
        """
        Returns a dictionary containing the settings for the given ini file
        section.

        :param str section: ini file section.
        """
        return dict(self.__ini_parser.items(section))

    def get_setting(self, section, key):
        """
        Returns the specified setting from the given ini file section.

        :param str section: ini file section.
        :param str key: key to look up in the section.
        """
        return self.__ini_parser.get(section, key)

    def has_setting(self, section, key):
        """
        Checks if the specified ini file section has a setting with the given
        name.
        """
        return self.__ini_parser.has_option(section, key)

    def get_app_url(self):
        """
        Returns the application URL as defined in the INI section
        `server:main`. Defaults to `http://0.0.0.0/6543`.
        """
        section = 'server:main'
        if self.has_setting(section, 'host'):
            host = self.get_setting(section, 'host')
        else:
            host = '0.0.0.0'
        if self.has_setting(section, 'port'):
            port = int(self.get_setting(section, 'port'))
        else:
            port = 6543
        return 'http://%s:%d' % (host, port)

    @classmethod
    def __check_ini_file(cls):
        if cls.__ini_parser is None:
            cls.__ini_parser = cls.__make_parser()
            if cls.ini_file_path is None:
                # Initialize with empty ini file.
                strm = NativeIO('[DEFAULT]')
                cls.__ini_parser.readfp(strm)
            else:
                cls.__ini_parser.read(cls.ini_file_path)
        return cls.__ini_parser

    @classmethod
    def __make_parser(cls):
        if not cls.ini_file_path is None:
            defaults = {'here':os.path.dirname(cls.ini_file_path)}
        else:
            defaults = {}
        return configparser.SafeConfigParser(defaults=defaults)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 23, 2012.
"""
from everest.resources.utils import get_root_collection
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityGrandchild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity

__docformat__ = 'reStructuredText en'
__all__ = ['create_collection',
           'create_entity',
           ]


def create_entity(entity_id=0, entity_text=None):
    my_entity = MyEntity(text=entity_text)
    my_entity.id = entity_id
    my_entity_parent = MyEntityParent()
    my_entity_parent.id = entity_id
    my_entity.parent = my_entity_parent
    my_entity_child = MyEntityChild()
    my_entity_child.id = entity_id
    my_entity.children.append(my_entity_child)
    my_entity_grandchild = MyEntityGrandchild()
    my_entity_grandchild.id = entity_id
    my_entity_child.children.append(my_entity_grandchild)
    # If we run with the SQLAlchemy backend, the back references are populated
    # automatically.
    if my_entity_child.parent is None:
        my_entity_child.parent = my_entity
    if my_entity_grandchild.parent is None:
        my_entity_grandchild.parent = my_entity_child
    return my_entity


def create_collection(entity_id1=0, entity_id2=1):
    my_entity1 = create_entity(entity_id=None, entity_text='foo0')
    my_entity2 = create_entity(entity_id=None, entity_text='too1')
    coll = get_root_collection(IMyEntity)
    coll.create_member(my_entity1)
    coll.create_member(my_entity2)
    my_entity1.id = entity_id1
    my_entity1.parent.id = entity_id1
    my_entity1.children[0].id = entity_id1
    my_entity1.children[0].children[0].id = entity_id1
    my_entity2.id = entity_id2
    my_entity2.parent.id = entity_id2
    my_entity2.children[0].id = entity_id2
    my_entity2.children[0].children[0].id = entity_id2
    return coll



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Package initialization file.

Created on Dec 1, 2011.
"""
from everest.configuration import Configurator
from everest.repositories.interfaces import IRepositoryManager
from everest.root import RootFactory


def app_factory(global_settings, **local_settings): # pylint: disable=W0613
    config = Configurator()
    config.setup_registry(settings=local_settings,
                          root_factory=RootFactory())
#    reg = get_current_registry()
#    repo_mgr = reg.getUtility(IRepositoryManager)
#    repo_mgr.initialize_all()
    return config.make_wsgi_app()



"""
Repository for the relational database (rdb) backend.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from sqlalchemy.engine import create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.pool import StaticPool

from everest.repositories.base import Repository
from everest.repositories.rdb.aggregate import RdbAggregate
from everest.repositories.rdb.querying import OptimizedCountingQuery
from everest.repositories.rdb.querying import SimpleCountingQuery
from everest.repositories.rdb.session import RdbSessionFactory
from everest.repositories.rdb.utils import empty_metadata
from everest.repositories.rdb.utils import get_metadata
from everest.repositories.rdb.utils import is_metadata_initialized
from everest.repositories.rdb.utils import map_system_entities
from everest.repositories.rdb.utils import reset_metadata
from everest.repositories.rdb.utils import set_metadata
from everest.repositories.utils import get_engine
from everest.repositories.utils import is_engine_initialized
from everest.repositories.utils import set_engine


__docformat__ = 'reStructuredText en'
__all__ = ['RdbRepository',
           ]


class RdbRepository(Repository):
    """
    Repository connected to a relational database backend (through an ORM).
    """
    _configurables = Repository._configurables \
                     + ['db_string', 'metadata_factory']

    def __init__(self, name, aggregate_class=None,
                 autoflush=True, join_transaction=True, autocommit=False):
        """
        :param autoflush: Sets the `autoflush` attribute.
        """
        if aggregate_class is None:
            aggregate_class = RdbAggregate
        Repository.__init__(self, name, aggregate_class,
                            join_transaction=join_transaction,
                            autocommit=autocommit)
        #: Flag indicating if changes should be flushed to the treansaction
        #: automatically.
        self.autoflush = autoflush
        # Default to an in-memory sqlite DB.
        self.configure(db_string='sqlite://',
                       metadata_factory=empty_metadata)

    def _initialize(self):
        # Manages a RDB engine and a metadata instance for this repository.
        # Both are global objects that should only be created once per process
        # (for each RDB repository), hence we use a global object manager.
        if not is_engine_initialized(self.name):
            engine = self.__make_engine()
            set_engine(self.name, engine)
            # Bind the engine to the session factory and the metadata.
            self.session_factory.configure(bind=engine)
        else:
            engine = get_engine(self.name)
        if not is_metadata_initialized(self.name):
            md_fac = self._config['metadata_factory']
            if self._config.get('messaging_enable', False):
                # Wrap the metadata callback to also call the mapping
                # function for system entities.
                reset_on_start = \
                    self._config.get('messaging_reset_on_start', False)
                def wrapper(engine, reset_on_start=reset_on_start):
                    metadata = md_fac(engine)
                    map_system_entities(engine, metadata, reset_on_start)
                    return metadata
                metadata = wrapper(engine)
            else:
                metadata = md_fac(engine)
            set_metadata(self.name, metadata)
        else:
            metadata = get_metadata(self.name)
        metadata.bind = engine

    def _reset(self):
        # It is safe to keep the engine around, even for complex unit test
        # scenarios; the metadata, however, might change and need to be
        # removed.
        if is_metadata_initialized(self.name):
            reset_metadata()

    def _make_session_factory(self):
        engine = get_engine(self.name)
        query_class = self.__check_query_class(engine)
        return RdbSessionFactory(self, query_class)

    def __make_engine(self):
        db_string = self._config['db_string']
        if db_string.startswith('sqlite://'):
            # Enable connection sharing across threads for pysqlite.
            kw = {'poolclass':StaticPool,
                  'connect_args':{'check_same_thread':False}
                  }
        else:
            kw = {} # pragma: no cover
        return create_engine(db_string, **kw)

    def __check_query_class(self, engine):
        # We check if the backend supports windowing for optimized counting.
        conn = engine.connect()
        try:
            conn.execute("select count(1) over()")
        except OperationalError:
            query_class = SimpleCountingQuery
        else:
            query_class = OptimizedCountingQuery # pragma: no cover
        return query_class
